From lrust.lifetime Require Export primitive.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import csum auth excl frac gmap agree gset.
Set Default Proof Using "Type".

Section accessors.
Context `{!invG Σ, !lftG Lat E0 Σ}.
Implicit Types κ : lft.

(* Internal helper lemmas *)
Lemma bor_open_internal E P κ s Pb q Vκ Vbor Vtok :
  ↑borN ⊆ E →
  slice borN s P -∗ ▷ lft_bor_alive κ Pb Vκ -∗
  idx_bor_own (κ, s) Vbor -∗ (q).[κ] Vtok ={E}=∗
    ∃ V, ⌜V ⊑ Vbor⌝ ∗ ▷ lft_bor_alive κ Pb Vκ ∗
         own_bor κ (◯ {[s := Excl (Bor_open q Vtok V)]}) ∗ ▷ P V.
Proof.
  iIntros (?) "Hslice Halive Hbor Htok". rewrite /lft_bor_alive /idx_bor_own.
  iDestruct "Halive" as (B) "[[Hbox >Hown] HB]". iDestruct "Hbor" as (V') "[% Hbor]".
  iDestruct (own_bor_valid_2 with "Hown Hbor")
    as %[EQB%to_borUR_included _]%auth_both_valid.
  iMod (slice_empty _ _ true with "Hslice Hbox") as "[HP Hbox]".
    solve_ndisj. by rewrite lookup_fmap EQB.
  rewrite -(fmap_insert bor_filled _ _ (Bor_open q Vtok V')).
  iMod (own_bor_update_2 with "Hown Hbor") as "[Hown Hbor]". apply auth_update.
  { eapply singleton_local_update. by rewrite lookup_fmap EQB.
    by apply (exclusive_local_update _ (Excl (Bor_open q Vtok V'))). }
  iExists V'. rewrite -fmap_insert. iFrame "∗%". iExists _. iFrame.
  rewrite -insert_delete big_sepM_insert ?lookup_delete // big_sepM_delete //=.
  iIntros "/= !>". iNext. iDestruct "HB" as "[?$]". iFrame.
Qed.

Lemma bor_close_internal E P κ s Pb q Vκ V Vtok Vtok' :
  ↑borN ⊆ E → Vtok ⊑ Vtok' →
  slice borN s P -∗ ▷ lft_bor_alive κ Pb Vκ -∗
  own_bor κ (◯ {[s := Excl (Bor_open q Vtok V)]}) -∗ ▷ P (Vtok' ⊔ V) ={E}=∗
    ▷ lft_bor_alive κ Pb (Vtok' ⊔ Vκ) ∗ idx_bor_own (κ, s) (Vtok' ⊔ V) ∗ (q).[κ] Vtok'.
Proof.
  iIntros (? HVV') "Hslice Halive Hbor HP". unfold lft_bor_alive, idx_bor_own.
  iDestruct "Halive" as (B) "[[Hbox >Hown] HB]".
  iDestruct (own_bor_valid_2 with "Hown Hbor")
    as %[EQB%to_borUR_included _]%auth_both_valid.
  iMod (slice_fill _ _ true with "Hslice HP Hbox") as "Hbox".
    solve_ndisj. by rewrite lookup_fmap EQB.
  rewrite -(fmap_insert bor_filled _ _ (Bor_in (Vtok' ⊔ V))).
  iMod (own_bor_update_2 with "Hown Hbor") as "Hbor". apply auth_update.
  { eapply singleton_local_update. by rewrite lookup_fmap EQB.
    by apply (exclusive_local_update _ (Excl (Bor_in (Vtok' ⊔ V)))). }
  rewrite own_bor_op -fmap_insert /=. iDestruct "Hbor" as "[Hown Hbor]".
  rewrite big_sepM_delete //. simpl. setoid_rewrite HVV'.
  iDestruct "HB" as "[>[% $] HB]". iSplitR "Hbor"; last by iExists (Vtok' ⊔ V); auto.
  iExists _. iFrame "Hbox Hown".
  rewrite -insert_delete big_sepM_insert ?lookup_delete //=. iFrame.
  iPureIntro. by f_equiv.
Qed.

Record atomic_map_extension (A A' : gmap atomic_lft (bool * Lat)) κ V : Prop :=
  { ame_lookup_fst Λ : fst <$> A' !! Λ = fst <$> A !! Λ;
    ame_hv κ Vκ : lft_has_view A κ Vκ → lft_has_view A' κ Vκ;
    ame_hv_κ Vκ : lft_has_view A κ Vκ → @lft_has_view Lat A' κ (V ⊔ Vκ) }.

Lemma extend_lft_view A κ q V :
  (q).[κ] V -∗ own_alft_auth A ==∗
  ∃ A', (q).[κ] V ∗ own_alft_auth A' ∗ ⌜atomic_map_extension A A' κ V⌝.
Proof.
  iIntros "Htok HA". rewrite /lft_tok /= . setoid_rewrite ->monPred_at_big_sepMS.
  iDestruct (big_opMS_gen_proper (A := atomic_lft)
    (λ P Q, ∀ A, P -∗ own_alft_auth A ==∗
         ∃ A', own_alft_auth A' ∗ Q ∗ ⌜∀ Λ, fst <$> A' !! Λ = fst <$> A !! Λ⌝ ∗
                    ⌜∀ κ Vκ, lft_has_view A κ Vκ → lft_has_view A' κ Vκ⌝)
    _ (λ Λ, own alft_name (◯ {[Λ := Cinl (q, to_latT V)]}))%I with "Htok")
    as "Htok".
  - intros ?. auto 10 with iFrame.
  - clear. iIntros (P1 P2 HP Q1 Q2 HQ A) "[HP1 HQ1] HA".
    iMod (HP with "HP1 HA") as (A') "(HA & HP2 & H)". iDestruct "H" as %[Hf1 Hhv1].
    iMod (HQ with "HQ1 HA") as (A'') "(HA & HQ2 & H)". iDestruct "H" as %[Hf2 Hhv2].
    iModIntro. iExists _. iFrame. iSplit; iIntros "!% **";
    [by rewrite Hf2 Hf1|by apply Hhv2, Hhv1].
  - clear. iIntros (Λ HΛ A) "Htok HA". iDestruct "Htok" as (V') "[% Htok]".
    iDestruct (own_valid_2 with "HA Htok") as
        %[(s0 & Hv & Hincl)%singleton_included _]%auth_both_valid.
    revert Hv Hincl. rewrite lookup_fmap. case HA: (A!!Λ)=>[[b V'']|]; [|by inversion 1].
    move=> <- /Some_included Hincl. destruct b; last first.
    { case Hincl=>[|/csum_included]; [by inversion 1|]. naive_solver. }
    iExists (<[Λ:=(true, V ⊔ V'')]>A).
    iMod (own_update_2 with "HA Htok") as "[$$]".
    { rewrite /to_alftUR fmap_insert. eapply auth_update, singleton_local_update.
      { by rewrite lookup_fmap HA. }
      apply csum_local_update_l, prod_local_update=>//=.
      etrans. apply (op_local_update_discrete _ _ (to_latT V))=>//.
      by rewrite !lat_op_join' (lat_le_join_l V V'). }
    iPureIntro; split.
    + intros Λ'. destruct (decide (Λ' = Λ)) as [->|?].
       by rewrite lookup_insert HA. by rewrite lookup_insert_ne.
    + intros κ' Vκ' Hκ' Λ' HΛ'. destruct (decide (Λ' = Λ)) as [->|?].
      * rewrite lookup_insert /=. specialize (Hκ' Λ HΛ'). rewrite HA /= in Hκ'.
        solve_lat.
      * rewrite lookup_insert_ne //. by apply Hκ'.
  - iMod ("Htok" with "HA") as (A') "(HA' & Htok & % &#Hhv)". iExists A'.
    iDestruct "Hhv" as %Hhv.
    iAssert (⌜lft_has_view A' κ V⌝)%I as %HA'V.
    { iIntros (Λ HΛ). iDestruct (@big_sepMS_elem_of with "Htok") as "Htok"=>//.
      iDestruct (own_valid_2 with "HA' Htok")
        as %[(s & HA' & [Hs|Hs]%Some_included)%singleton_included _]%auth_both_valid.
      - setoid_subst. rewrite lookup_fmap /to_lft_stateR in HA'.
        destruct (A'!!Λ) as [[[] V']|]=>/=.
        + apply (inj Some), (inj Cinl), (inj2 pair), proj2, (inj to_latT) in HA'=>//.
          by setoid_subst.
        + apply (inj Some) in HA'. inversion HA'.
        + inversion HA'.
      - rewrite lookup_fmap in HA'. destruct (A'!!Λ) as [[]|]; [|by inversion HA'].
        apply (inj Some) in HA'. setoid_subst. apply csum_included in Hs.
        destruct b, Hs as [[=]|[(?&?&[=]&[=]&[_ Hincl%latT_included]%prod_included)|
          (?&?&[=]&?)]]. auto with subst. }
    iFrame. iModIntro. iSplit.
    + iApply (@big_sepMS_mono with "Htok")=>//= Λ _. iIntros "?".
      iExists _. by iFrame.
    + iPureIntro. split=>// Vκ HVκ Λ HΛ. specialize (Hhv κ Vκ HVκ Λ HΛ).
      specialize (HA'V Λ HΛ). destruct (A' !! Λ)=>//. simpl in *. solve_lat.
Qed.

Lemma ame_lft_alive_in A A' κ V κ' :
  atomic_map_extension A A' κ V → lft_alive_in A κ' → lft_alive_in A' κ'.
Proof. intros ????. erewrite (ame_lookup_fst A A')=>//. auto. Qed.
Lemma ame_lft_dead_in A A' κ V κ' :
  atomic_map_extension A A' κ V → lft_dead_in A κ' → lft_dead_in A' κ'.
Proof. intros ?(?&?&?). eexists _. erewrite (ame_lookup_fst A A')=>//. Qed.
Lemma ame_lft_inv A A' κ V κ' :
  atomic_map_extension A A' κ V → lft_inv A κ' -∗ lft_inv A' κ'.
Proof.
  intros ?. unfold lft_inv. iDestruct 1 as (Vκ') "[% H]". iExists Vκ'.
  iSplit; [by eauto using ame_hv|].
  iDestruct "H" as "[[?%]|[?%]]"; [iLeft|iRight]; iFrame; iPureIntro.
    by eapply ame_lft_alive_in. by eapply ame_lft_dead_in.
Qed.

(** Internal accessors *)
Lemma idx_bor_acc_internal E q κ s P Vbor Vtok :
  ↑borN ⊆ E → ↑mgmtN ⊆ E →
  lft_ctx -∗ slice borN s P -∗ idx_bor_own (κ, s) Vbor -∗ q.[κ] Vtok ={E}=∗
    ∃ V, ⌜V ⊑ Vbor⌝ ∗ ▷ P V ∗
      (∀ Vtok', ⌜Vtok ⊑ Vtok'⌝ -∗ ▷ P (Vtok' ⊔ V) ={E}=∗
                   idx_bor_own (κ, s) (Vtok' ⊔ V) ∗ q.[κ] Vtok').
Proof.
  iIntros (HE1 HE2) "#LFT #Hs Hbor Htok".
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  iAssert ⌜κ ∈ dom (gset _) I⌝%I as %Hκ'.
  { unfold idx_bor_own. iDestruct "Hbor" as (Vb) "[% Hbor]".
    iApply (own_bor_auth with "HI Hbor"). }
  rewrite big_sepS_later big_sepS_elem_of_acc //. rewrite [lft_inv _ κ]/lft_inv.
  iDestruct "Hinv" as "[Hinv Hclose']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iMod (lft_dead_in_tok with "HA") as "[_ H†]". done.
    iDestruct "H†" as (V) "H†".
    set (V' := Vtok ⊔ V).
    iDestruct (lft_tok_dead $! V' with "Htok H†") as "[]". }
  rewrite [in _ Vκ]lft_inv_alive_unfold.
  iDestruct "Hinv" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iMod (bor_open_internal with "Hs Halive Hbor Htok") as (V) "(% & Halive & Hf & HP)".
    solve_ndisj.
  iExists V. iFrame "HP %".
  iMod ("Hclose" with "[-Hf]") as "_".
  { iExists _, _. iFrame. rewrite big_sepS_later. iApply "Hclose'".
    iExists Vκ. iFrame "%". iLeft. rewrite lft_inv_alive_unfold. iExists _, _. iFrame. }
  iModIntro. iIntros (Vtok' HVtok') "HP".
  clear -HE1 HE2 HVtok'. iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  iDestruct (own_bor_auth with "HI Hf") as %Hκ'.
  rewrite big_sepS_later big_sepS_delete //.
  rewrite [lft_inv _ κ]/lft_inv /lft_inv_dead /lft_bor_dead.
  iDestruct "Hinv" as "[Hinv Hclose']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iDestruct "Hinv" as (??) "(_ & Hinv & _)". iDestruct "Hinv" as (B ?) "[>Hinv _]".
    iDestruct (own_bor_valid_2 with "Hinv Hf")
      as %[(_ & <- & INCL%option_included)%singleton_included _]%auth_both_valid.
    exfalso. destruct INCL as [[=]|(? & ? & [=<-] & Hi & Ho)].
    rewrite lookup_fmap in Hi. destruct (B !! s); inversion Hi; subst; clear Hi.
    by destruct Ho as [?%(inj Excl)|?%(exclusive_included _)]. }
  rewrite lft_inv_alive_unfold.
  iDestruct "Hinv" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iMod (bor_close_internal with "Hs Halive Hf HP") as "(Halive & $ & Htok)".
    solve_ndisj. done.
  iMod (extend_lft_view with "Htok HA") as (A') "($ & HA' & %)".
  iApply "Hclose". iExists _, _. iFrame.
  rewrite big_sepS_later (big_sepS_delete _ (dom _ I)) //. iSplitR "Hclose'".
  - unfold lft_inv. iExists (Vtok' ⊔ Vκ). iSplit; [by eauto using ame_hv_κ|]. iLeft. iSplit.
    + rewrite lft_inv_alive_unfold. iExists _, _. iFrame.
    + eauto using ame_lft_alive_in.
  - iApply (@big_sepS_impl with "[$Hclose']"). iIntros "!# * % ?". by iApply ame_lft_inv.
Qed.

(** Indexed borrow *)
Lemma idx_bor_acc E q κ i P Vtok Vbor:
  ↑lftN ⊆ E →
  lft_ctx -∗ &{κ,i}P Vtok -∗ idx_bor_own i Vbor -∗ q.[κ] Vtok ={E}=∗
    ∃ V, ⌜V ⊑ Vbor ⊔ Vtok⌝ ∗ ▷ P V ∗
          (∀ Vtok', ⌜Vtok ⊑ Vtok'⌝ -∗ ▷ P (Vtok' ⊔ V)
                    ={E}=∗ idx_bor_own i (Vtok' ⊔ V) ∗ q.[κ] Vtok').
Proof.
  unfold idx_bor. iIntros (HE) "#LFT [#Hle #Hs] Hbor Htok".
  iDestruct "Hs" as (P') "[#HPP' Hs]". destruct i as [κ' s].
  iMod (lft_incl_acc with "Hle Htok") as (q') "[Htok Hclose]". solve_ndisj.
  iMod (idx_bor_acc_internal with "LFT Hs Hbor Htok") as (V) "(% & HP & Hclose')".
    solve_ndisj. solve_ndisj.
  iExists (V ⊔ Vtok). iSplitR; [by iPureIntro; f_equiv|]. iModIntro. iSplitL "HP".
  - iNext. by iApply "HPP'".
  - iIntros (Vtok' ?) "HP". iMod ("Hclose'" with "[//] [HP]") as "[$ Htok]".
    + iApply "HPP'". iFrame.
    + by iMod ("Hclose" with "Htok").
Qed.

(** Basic borrows  *)
Lemma add_vs Pb Pb' P Q Pi κ :
  ▷ (Pb ≡ (P ∗ Pb')) -∗ lft_vs κ Pb Pi -∗ (▷ Q -∗ <subj> [†κ] ={E0}=∗ ▷ P) -∗
  lft_vs κ (Q ∗ Pb') Pi.
Proof.
  iIntros "#HEQ Hvs HvsQ". iApply (lft_vs_cons with "[-Hvs] Hvs").
  iIntros "$ [HQ HPb'] #H†".
  iApply fupd_mask_mono; [|iMod ("HvsQ" with "HQ H†") as "HP"].
    set_solver.
  iModIntro. iNext. iRewrite "HEQ". iFrame.
Qed.

(* TODO : the dead tokens that we get in the "consequence" VS are
   under the objective modality. This is because we have to execute
   the VS at a view which is smaller that all the views of the atomic
   liftimes of κ. But such a dead token does not necessarily exist.

   Actually, we could make this to work by havng a CMRA that maintain
   a view per lifetime for dead tokens (as opposed to a view per
   atomic lifetimes for alive tokens). But this CMRA would be
   significantly complex, because it would need to maintain the
   monotonicity of this map wrt static inclusion, and, moreover, the
   backward direction of lft_dead_or would break. *)

Lemma bor_acc_strong E q κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ∃ κ', κ ⊑ κ' ∗ ▷ P ∗
    ∀ Q, ▷ (▷ Q -∗ <subj> [†κ'] ={E0}=∗ ▷ P) -∗ ▷ Q ={E}=∗ &{κ'}Q ∗ q.[κ].
Proof.
  unfold bor, raw_bor. iIntros (HE) "#LFT Hbor Htok".
  iDestruct "Hbor" as (κ'') "(#Hle & Htmp)". iDestruct "Htmp" as (s'') "(Hbor & #Hs)".
  iDestruct "Hs" as (P') "[#HPP' Hs]".
  iMod (lft_incl_acc with "Hle Htok") as (q') "[Htok Hclose]". solve_ndisj.
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose'".
  iAssert (⌜κ'' ∈ dom (gset _) I⌝)%I as %Hκ'.
  { unfold idx_bor_own. iDestruct "Hbor" as (Vb) "[#HVb Hbor]".
    rewrite -embed_pure. iApply (own_bor_auth with "HI Hbor"). }
  rewrite big_sepS_later big_sepS_elem_of_acc //. rewrite [lft_inv _ κ'']/lft_inv.
  iDestruct "Hinv" as "[Hinv Hclose'']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iMod (lft_dead_in_tok with "HA") as "[_ H†]". done.
    iDestruct (monPred_in_intro with "Htok") as (Vtok) "[_ Htok]".
    iDestruct "H†" as (V) "H†". iAssert ⎡False⎤%I as %[]. iModIntro.
    set (V' := Vtok ⊔ V).
    iApply (lft_tok_dead $! V' with "Htok H†"). }
  rewrite [in _ Vκ]lft_inv_alive_unfold.
  iDestruct "Hinv" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iCombine "Hbor Htok" as "HborHtok".
  iDestruct (monPred_in_intro with "HborHtok") as (V) "(#HV & Hbor & Htok)".
  iMod (bor_open_internal with "Hs Halive Hbor Htok")
    as (V0 HVV0) "(Halive & Hbor & HP')". solve_ndisj.
  iAssert (▷ P)%I with "[HP']" as "$".
  { iNext. iApply "HPP'". iApply monPred_in_elim. done. iFrame. }
  iMod ("Hclose'" with "[-Hbor Hclose]") as "_".
  { iExists _, _. iFrame. rewrite big_sepS_later. iApply "Hclose''".
    iExists _. iFrame "%". iLeft. rewrite lft_inv_alive_unfold. iExists _, _. iFrame. }
  iExists κ''. iFrame "#". iModIntro. iIntros (Q) "HvsQ HQ". clear -HE HVV0.
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose'".
  iDestruct (own_bor_auth with "HI Hbor") as %Hκ'.
  rewrite big_sepS_later big_sepS_delete //.
  rewrite [lft_inv _ κ'']/lft_inv /lft_inv_dead /lft_bor_dead.
  iDestruct "Hinv" as "[Hinv Hclose'']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iDestruct "Hinv" as (??) "(_ & Hinv & _)". iDestruct "Hinv" as (B ?) "[>Hinv _]".
    iDestruct (own_bor_valid_2 with "Hinv Hbor")
      as %[(_ & <- & INCL%option_included)%singleton_included _]%auth_both_valid.
    exfalso. destruct INCL as [[=]|(? & ? & [=<-] & Hi & Ho)].
    rewrite lookup_fmap /= in Hi. destruct (B !! s''); inversion Hi; subst; clear Hi.
    by destruct Ho as [?%(inj Excl)|?%(exclusive_included _)]. }
  rewrite lft_inv_alive_unfold /lft_bor_alive.
  iDestruct "Hinv" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iDestruct "Halive" as (B) "[[Hbox >Hown] HB]".
  iDestruct (own_bor_valid_2 with "Hown Hbor")
    as %[EQB%to_borUR_included _]%auth_both_valid.
  iDestruct (slice_delete_empty _ true with "Hs Hbox") as (Pb') "[EQ Hbox]".
    by rewrite lookup_fmap EQB.
  iMod (slice_insert_empty _ _ true _ Q with "Hbox") as (j) "(% & #Hs' & Hbox)".
  iMod (own_bor_update_2 with "Hown Hbor") as "Hown".
  { apply auth_update. etrans.
    - apply delete_singleton_local_update, _.
    - apply (alloc_singleton_local_update _ j (Excl (Bor_open q' V V0))); last done.
      rewrite -fmap_delete lookup_fmap fmap_None
              -fmap_None -lookup_fmap fmap_delete //. }
  iDestruct "Hown" as "[Hown Hbor]".
  iAssert (▷ (▷ Q -∗ <subj> [†κ''] ={E0}=∗ ▷ P'))%I with "[HvsQ]" as "HvsQ".
  { iNext. iIntros "HQ H†". iApply "HPP'". iApply ("HvsQ" with "HQ H†"). }
  iCombine "HV HQ HvsQ" as "HH".
  iDestruct (monPred_in_intro with "HH") as (V') "(#HV' & % & HQ & HvsQ)".
  iMod (bor_close_internal with "Hs' [Hbox Hown HB] Hbor [HQ]")
    as "(Halive & Hbor & Htok)". solve_ndisj. done.
  { rewrite -!fmap_delete -fmap_insert -(fmap_insert _ _ _ (Bor_open q' V V0))
            /lft_bor_alive. iExists _. iFrame "Hbox Hown".
    rewrite big_sepM_insert. by rewrite big_sepM_delete.
    by rewrite -fmap_None -lookup_fmap fmap_delete. }
  { rewrite -(_ : V' ⊑ V' ⊔ V0) //. solve_lat. }
  iMod (extend_lft_view with "Htok HA") as (A') "(Htok & HA' & %)".
  iMod ("Hclose'" with "[-Hbor Htok Hclose]") as "_".
  { iModIntro. iModIntro. iExists _, _. iFrame.
    rewrite (big_sepS_delete _ (dom _ I)) //. iSplitR "Hclose''".
    - unfold lft_inv. iExists (V' ⊔ Vκ). iSplit; [by eauto using ame_hv_κ|]. iLeft. iSplit.
      + rewrite lft_inv_alive_unfold /lft_bor_alive. iExists _, _.
        iFrame "Hinh". iSplitL "Halive"; [iDestruct "Halive" as (?) "?"; by iExists _|].
        iApply (add_vs with "EQ Hvs"). iFrame.
      + eauto using ame_lft_alive_in.
    - iApply (@big_sepS_impl (uPredI _) with "[$Hclose'']"). iIntros "!# % % ?".
        by iApply ame_lft_inv. }
  iDestruct (monPred_in_elim with "HV' Htok") as "Htok".
  iMod ("Hclose" with "Htok") as "$". iExists κ''. iModIntro.
  iSplit; first by iApply lft_incl_refl. iExists j.
  iDestruct (monPred_in_elim with "[] Hbor") as "$".
  { rewrite (_ : V' ⊔ V0 ⊑ V') //. solve_lat. }
  iExists _. iFrame "Hs'". by rewrite -bi.iff_refl.
Qed.

Lemma bor_acc_atomic_strong E κ P:
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P ={E,E∖↑lftN}=∗
    (∃ P' κ', κ ⊑ κ' ∗ ▷ (⎡P'⎤ ∧ P) ∗
       ∀ Q,
         ▷ (▷ Q -∗ <subj> [†κ'] ={E0}=∗ ▷ P) -∗
         ▷ (⎡P'⎤ ∧ Q)
         ={E∖↑lftN,E}=∗ &{κ'}Q) ∨
    (∃ κ', κ ⊑ κ' ∗ <subj> [†κ'] ∗ |={E∖↑lftN,E}=> True).
Proof.
  iIntros (HE) "#LFT Hbor". rewrite bor_unfold_idx /idx_bor.
  iDestruct "Hbor" as (i) "((#Hle & #Hs) & Hbor)".
  iDestruct "Hs" as (P') "[#HPP' Hs]".
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose'".
  iAssert (⌜i.1 ∈ dom (gset _) I⌝)%I as %Hκ'.
  { unfold idx_bor_own. iDestruct "Hbor" as (Vb) "[#HVb Hbor]".
    by iDestruct (own_bor_auth with "HI Hbor") as "%". }
  rewrite big_sepS_later big_sepS_elem_of_acc //. rewrite [lft_inv _ (i.1)]/lft_inv.
  iDestruct "Hinv" as "[Hinv Hclose'']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iRight. iMod (lft_dead_in_tok with "HA") as "[HA #H†]". done.
    iMod ("Hclose'" with "[-Hbor]") as "_".
    - iExists _, _. iFrame. rewrite big_sepS_later. iApply "Hclose''".
      iExists _. iFrame "%". iRight. by iFrame.
    - iExists (i.1). rewrite monPred_subjectively_unfold /=. iFrame "#".
      rewrite bi.True_emp. iApply fupd_intro_mask'. solve_ndisj. }
  iLeft. unfold lft_bor_alive, idx_bor_own. iDestruct "Hbor" as (Vb) "[#HVb Hbor]".
  iExists (P' Vb), (i.1). iFrame "#". rewrite [in _ Vκ]lft_inv_alive_unfold /lft_bor_alive.
  iDestruct "Hinv" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iDestruct "Halive" as (B) "[[Hbox >Hown] HB]".
  iDestruct (own_bor_valid_2 with "Hown Hbor")
      as %[EQB%to_borUR_included _]%auth_both_valid.
  iMod (slice_delete_full _ _ true with "Hs Hbox") as (Pb') "(HP' & EQ & Hbox)".
    solve_ndisj. by rewrite lookup_fmap EQB.
  iApply fupd_frame_l. iSplitL "HP'".
  { iNext. iSplit; [done|]. iApply "HPP'". by iApply monPred_in_elim. }
  iMod fupd_intro_mask' as "Hclose"; [|iModIntro; iIntros (Q) "HvsQ HQ"].
    solve_ndisj.
  iMod "Hclose" as "_".
  iAssert (▷⌜Vb ⊑ Vκ⌝)%I as "#>%".
  { rewrite monPred_at_big_sepM. rewrite -embed_pure. iNext. iModIntro.
    rewrite (big_sepM_lookup _ _ _ _ EQB) /= monPred_at_in //. }
  iCombine "HQ HvsQ HPP'" as "HH".
  iDestruct (monPred_in_intro with "HH") as (V) "(#HV & HQ & HvsQ & #HPP'V)".
  set (Q' := ((monPred_in V ∨ ⎡P' Vb⎤) ∧ (monPred_in V → Q))%I).
  iMod (slice_insert_full _ _ true _ _ Q' Vb with "[HQ] Hbox")
    as (j) "(% & #Hs' & Hbox)". solve_ndisj.
  { rewrite /Q' /=. iSplit.
    - iDestruct "HQ" as "[? _]". iNext. by iRight.
    - iDestruct "HQ" as "[_ ?]". iIntros (???). iFrame. }
  iMod (own_bor_update_2 with "Hown Hbor") as "Hown".
  { apply auth_update. etrans.
    - apply delete_singleton_local_update, _.
    - apply (alloc_singleton_local_update _ j (Excl (Bor_in Vb))); last done.
      rewrite -fmap_delete lookup_fmap fmap_None
              -fmap_None -lookup_fmap fmap_delete //. }
  rewrite own_bor_op. iDestruct "Hown" as "[H● H◯]".
  iMod ("Hclose'" with "[- H◯]") as "_"; last first.
  { iModIntro. unfold bor, raw_bor, idx_bor_own.
    iExists (i.1). iSplit; first by iApply lft_incl_refl.
    iExists j. iSplitL "H◯"; [by iExists _; iFrame|].
    iExists _. iFrame "#". iIntros "!>!#". iSplit; iIntros "H /=".
    - by iApply "H".
    - rewrite /Q'. auto. }
  iModIntro ⎡_⎤%I. iDestruct (add_vs _ _ _ Q' _ _ with "EQ Hvs [HvsQ]") as "Hvs".
  { iNext. iIntros "%% HQ" (Vκ') "% H†". rewrite bi.and_or_r.
    iDestruct "HQ" as "[[>% HQ]|[$ _]]"; [|done].
    iSpecialize ("HQ" with "[//]"). rewrite -monPred_at_later.
    iApply "HPP'V". iApply ("HvsQ" with "HQ H†"). }
  iExists _, _. iFrame. rewrite big_sepS_later. iApply "Hclose''".
  iExists _. iFrame "%". iLeft. rewrite lft_inv_alive_unfold. iExists _, _. iFrame.
  iNext. rewrite -!fmap_delete -fmap_insert -(fmap_insert _ _ _ (Bor_in Vb)).
  rewrite /lft_bor_alive. iExists _. iFrame "Hbox H●".
  rewrite big_sepM_insert. by rewrite big_sepM_delete.
    by rewrite -fmap_None -lookup_fmap fmap_delete.
Qed.
End accessors.
