From lrust.lifetime.model Require Export primitive.
From iris.algebra Require Import csum auth excl frac gmap agree gset.
From iris.proofmode Require Import tactics.
Set Default Proof Using "Type*".

Section faking.
Context `{!invG Σ, !lftG Lat E0 Σ, LatBottom Lat bot}.
Implicit Types κ : lft.

Lemma ilft_create A I κ :
  own_alft_auth A -∗ own_ilft_auth I -∗ ▷ ([∗ set] κ ∈ dom _ I, lft_inv A κ)
    ==∗ ∃ A' I', ⌜κ ∈ dom (gset _) I'⌝ ∗
      own_alft_auth A' ∗ own_ilft_auth I' ∗ ▷ ([∗ set] κ ∈ dom _ I', lft_inv A' κ).
Proof.
  iIntros "HA HI Hinv".
  destruct (decide (is_Some (I !! κ))) as [?|HIκ%eq_None_not_Some].
  { iModIntro. iExists A, I. iFrame. iPureIntro. by apply elem_of_dom. }
  iMod (own_alloc (● 0 ⋅ ◯ 0)) as (γcnt) "[Hcnt Hcnt']"; first by apply auth_both_valid.
  iMod (own_alloc ((● ∅ ⋅ ◯ ∅) :auth (gmap slice_name (excl (bor_state Lat)))))
    as (γbor) "[Hbor Hbor']"; first by apply auth_both_valid.
  iMod (own_alloc ((● ε) :auth (gset_disj slice_name)))
     as (γinh) "Hinh"; first by apply auth_auth_valid.
  set (γs := LftNames γbor γcnt γinh).
  iMod (own_update with "HI") as "[HI Hγs]".
  { apply auth_update_alloc,
      (alloc_singleton_local_update _ κ (to_agree γs)); last done.
    by rewrite lookup_fmap HIκ. }
  iDestruct "Hγs" as "#Hγs".
  iAssert (own_cnt κ (● 0)) with "[Hcnt]" as "Hcnt".
  { rewrite /own_cnt. iExists γs. by iFrame. }
  iAssert (own_cnt κ (◯ 0)) with "[Hcnt']" as "Hcnt'".
  { rewrite /own_cnt. iExists γs. by iFrame. }
  iAssert (∀ o, lft_inh κ o True)%I with "[Hinh]" as "Hinh".
  { iIntros (o). rewrite /lft_inh. iExists ∅. rewrite gset_to_gmap_empty.
    iSplitL; [|iApply box_alloc]. rewrite /own_inh. iExists γs. by iFrame. }
  iAssert (lft_inv_dead κ bot ∧ lft_inv_alive κ bot)%I
    with "[-HA HI Hinv]" as "Hdeadandalive".
  { iSplit.
    - rewrite /lft_inv_dead. iExists True%I, bot. iFrame "Hcnt".
      iSplit; [done|]. iSplitL "Hbor"; last by iApply "Hinh". rewrite /lft_bor_dead.
      iExists ∅, True%I. rewrite !fmap_empty.
      iSplitL "Hbor". iExists γs. by iFrame. iApply box_alloc.
    - rewrite lft_inv_alive_unfold. iExists True%I, True%I. iSplitL "Hbor".
      { rewrite /lft_bor_alive. iExists ∅.
        rewrite /to_borUR !fmap_empty big_sepM_empty monPred_at_emp (right_id ).
        iSplitR; [iApply box_alloc|]. rewrite /own_bor. iExists γs. by iFrame. }
      iSplitR "Hinh"; last by iApply "Hinh".
      rewrite lft_vs_unfold. iExists bot, 0. iSplit=>//. iFrame "Hcnt Hcnt'". auto. }
  set (A' := union_with (λ x _, Some x) A (gset_to_gmap (false, bot) (dom _ κ))).
  iMod (own_update _ _ (● to_alftUR A' ⋅ _) with "HA") as "[HA _]".
  { apply auth_update_alloc.
    assert (to_alftUR A'
      ≡ gset_to_gmap (Cinr (to_agree $ to_latT bot)) (dom _ κ ∖ dom _ A) ⋅ to_alftUR A) as ->.
    { intros Λ. rewrite lookup_op lookup_gset_to_gmap !lookup_fmap lookup_union_with
        lookup_gset_to_gmap.
      destruct (A !! Λ) eqn:EQ.
      - apply (elem_of_dom_2 (D:=gset atomic_lft)) in EQ.
        rewrite [X in _ ≡ X ⋅ _]option_guard_False; last set_solver. by destruct mguard.
      - apply (not_elem_of_dom (D:=gset atomic_lft)) in EQ.
        destruct (decide (Λ ∈ dom (gset _) κ)).
        + rewrite !option_guard_True; set_solver.
        + rewrite !option_guard_False; set_solver. }
    eapply op_local_update_discrete=>HA Λ. specialize (HA Λ).
    rewrite lookup_op lookup_gset_to_gmap !lookup_fmap.
    destruct (A !! Λ) eqn:EQ.
    - apply (elem_of_dom_2 (D:=gset atomic_lft)) in EQ.
      rewrite option_guard_False; [by destruct p as [[]?]|set_solver].
    - apply (not_elem_of_dom (D:=gset atomic_lft)) in EQ.
      destruct (decide (Λ ∈ dom (gset _) κ)).
      + rewrite !option_guard_True; [done|set_solver].
      + rewrite !option_guard_False; [done|set_solver]. }
  iExists A', (<[κ:=γs]> I). iFrame "HA". iModIntro.
  iSplit; first by rewrite elem_of_dom lookup_insert; eauto.
  rewrite /own_ilft_auth /to_ilftUR fmap_insert dom_insert_L. iFrame "HI".
  iNext. iApply @big_sepS_insert; first by apply not_elem_of_dom. iSplitR "Hinv".
  - destruct (lft_alive_or_dead_in A' κ) as [(Λ&?&HAΛ)|Haliveordead].
    + rewrite lookup_union_with lookup_gset_to_gmap option_guard_True in HAΛ;
        [by destruct (A!!Λ)|by apply gmultiset_elem_of_dom].
    + unfold lft_inv. iExists bot. iSplit; last first.
      { destruct Haliveordead.
        * iLeft. by iDestruct "Hdeadandalive" as "[_ $]".
        * iRight. by iDestruct "Hdeadandalive" as "[$ _]". }
      iPureIntro=>Λ?. rewrite lookup_union_with lookup_gset_to_gmap option_guard_True;
        [|by apply gmultiset_elem_of_dom]. destruct (A!!Λ); apply lat_bottom_sqsubseteq.
  - iApply (@big_sepS_impl with "[$Hinv]"). iIntros "!# * _ H". unfold lft_inv.
    iDestruct "H" as (Vκ) "[HV Hinv]". iExists Vκ. iSplit.
    + iDestruct "HV" as %HV. iPureIntro. intros Λ HΛ. specialize (HV Λ HΛ).
      rewrite lookup_union_with. by destruct (A !! Λ), (gset_to_gmap _ _ !! Λ).
    + iDestruct "Hinv" as "[[Hinv Hin]|[Hinv Hin]]".
      * iLeft. iFrame. iDestruct "Hin" as %Hin. iPureIntro.
        intros Λ HΛ. specialize (Hin Λ HΛ). rewrite lookup_union_with.
        by destruct (A !! Λ), (gset_to_gmap _ _ !! Λ).
      * iRight. iFrame. iDestruct "Hin" as %(Λ & HΛ & HA). iPureIntro.
        exists Λ. split; first done. rewrite lookup_union_with.
        by destruct (A !! Λ), (gset_to_gmap _ _ !! Λ).
Qed.

Lemma raw_bor_fake E κ P :
  ↑borN ⊆ E → ▷ ⎡lft_bor_dead κ⎤ ={E}=∗ ▷ ⎡lft_bor_dead κ⎤ ∗ raw_bor κ P.
Proof.
  iIntros (?). rewrite /lft_bor_dead. iDestruct 1 as (B Pinh) "[>HB● Hbox]".
  iMod (slice_insert_empty _ _ true _ P with "[Hbox]") as (γ) "(HB & #Hslice & Hbox)".
  { by iNext. }
  iDestruct "HB" as %HB. iMod (own_bor_update with "HB●") as "[HB● H◯]".
  { eapply auth_update_alloc,
      (alloc_singleton_local_update _ γ (Excl (Bor_in bot))); last done.
    move:HB. rewrite !lookup_fmap. by destruct (B !! γ). }
  rewrite /bor /raw_bor /idx_bor_own /=. iModIntro. iSplitR "H◯".
  - iExists (<[γ:=bot]> B), (P ∗ Pinh)%I. rewrite !fmap_insert. iNext. by iFrame.
  - iExists γ. iSplitL.
    + iDestruct (monPred_in_intro True%I with "[//]") as (V) "[? _]".
      iExists bot. iFrame. by rewrite [bot]lat_bottom_sqsubseteq.
    + iExists P. iIntros "{$Hslice} !> !# **". by iApply (bi.iff_refl True%I).
Qed.

Lemma raw_bor_fake' E κ P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ <subj> [†κ] ={E}=∗ raw_bor κ P.
Proof.
  iIntros (?) "#LFT H†". iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  iMod (ilft_create _ _ κ with "HA HI Hinv") as (A' I') "(Hκ & HA & HI & Hinv)".
  iDestruct "Hκ" as %Hκ. rewrite /lft_dead monPred_subjectively_unfold /=.
  iDestruct "H†" as (V0 Λ V'') "(% & % & #H†)".
  iDestruct (own_alft_auth_agree A' Λ (false, V'') with "HA H†") as %EQAΛ.
  rewrite big_sepS_elem_of_acc; [|done]. iDestruct "Hinv" as "[Hinv Hclose']".
  rewrite {1}/lft_inv; iDestruct "Hinv" as (Vκ) "[>%[[_ >HA']|[Hinv >%]]]".
  { iDestruct "HA'" as %HA'. specialize (HA' Λ).
    rewrite <-leibniz_equiv_iff, ->EQAΛ in HA'. naive_solver. }
  rewrite /lft_inv_dead; iDestruct "Hinv" as (Pinh Vinh) "(>% & Hdead & Hcnt & Hinh)".
  iMod (raw_bor_fake _ _ P with "[Hdead]") as "[Hdead Hbor]"; first solve_ndisj.
  { by iNext. }
  iFrame. iMod ("Hclose" with "[-]"); [|done]. iExists A', I'. iFrame. iNext.
  iApply "Hclose'". rewrite /lft_inv /lft_inv_dead. iExists Vκ.
  iFrame "%". iRight. iFrame "∗%". iExists _, _. by iSplit.
Qed.

Lemma bor_fake E κ P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ <subj> [†κ] ={E}=∗ &{κ}P.
Proof.
  iIntros (?) "#LFT H†". iMod (raw_bor_fake' with "LFT H†"); first done.
  iModIntro. unfold bor. iExists κ. iFrame. iApply lft_incl_refl.
Qed.
End faking.
