From iris.algebra Require Import auth.
From lrust.lifetime Require Import borrow accessors.
From iris.proofmode Require Import tactics.
Set Default Proof Using "Type".

(** Internal atomic persistent borrows. *)
(* They are like atomic persistent borrows, but use lftN as their
   namespace. They are essential to define fractured borrows.
   Contrary to the non-view lifetime logic, it seems like the view
   lifetime logic cannot define them outside of the logic by using
   fractionnal borrow tokens. The reason is that a fraction of the
   borrow token would not have a synchronizing (e.g., with tokens)
   accessor, which is needed to give the full ownership of the payload
   of the borrow. TODO: is there any way of taking this out of the
   internal lifetime logic?  *)

Definition in_atN : namespace := lftN .@ "in_at".

Definition in_at_bor {Σ Lat} `{invG Σ, lftG Lat E0 Σ} κ P :=
    (∃ P' i, (κ ⊑ (i.1))
               ∗ ▷ □ (P' ↔ P) ∗ ⎡slice borN (i.2) P'⎤
               ∗ ⎡inv in_atN (∃ V', idx_bor_own i V')⎤)%I.
Notation "&in_at{ κ }" := (in_at_bor κ) (format "&in_at{ κ }") : bi_scope.
Instance in_at_bor_params : Params (@in_at_bor) 5 := {}.

Section in_at.
Context `{invG Σ, lftG Lat E0 Σ}.

Global Instance in_at_bor_ne κ n : Proper (dist n ==> dist n) (in_at_bor κ).
Proof. solve_proper. Qed.
Global Instance in_at_bor_contractive κ : Contractive (in_at_bor κ).
Proof. solve_contractive. Qed.
Global Instance in_at_bor_proper : Proper ((≡) ==> (≡)) (in_at_bor κ).
Proof. solve_proper. Qed.
Global Instance in_at_bor_persistent : Persistent (&in_at{κ}P) := _.

Lemma in_at_bor_iff κ P P' :
  ▷ □ (P ↔ P') -∗ &in_at{κ}P -∗ &in_at{κ}P'.
Proof.
  iIntros "#HPP' H". iDestruct "H" as (P0 i) "#(? & HPP0 & ? & ?)".
  iExists P0, i. iFrame "#%". iNext. iIntros "!# **". iSplit; iIntros.
  { iApply "HPP'". by iApply "HPP0". }  { iApply "HPP0". by iApply "HPP'". }
Qed.

Lemma in_at_bor_share E κ P : &{κ}P ={E}=∗ &in_at{κ}P.
Proof.
  rewrite bor_unfold_idx /idx_bor. iDestruct 1 as (i) "[[? H] Hown]".
  iDestruct "H" as (P') "[HPP' Hs]". iExists P', i. iFrame "∗%".
  iMod (inv_alloc with "[Hown]") as "$"; [|done]. iNext.
  iDestruct (monPred_subjectively_intro with "Hown") as "Hown".
  by rewrite monPred_subjectively_unfold.
Qed.

Lemma in_at_bor_acc E q κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &in_at{κ}P -∗ q.[κ] ={E,E∖↑lftN}=∗ ∃ Vb,
    (monPred_in Vb → ▷ P) ∗ ((monPred_in Vb → ▷ P) ={E∖↑lftN,E}=∗ q.[κ]).
Proof.
  iIntros (?) "#LFT Hb Htok". iDestruct "Hb" as (P' [κ' s]) "#(? & HPP' & Hs & ?)".
  iMod (lft_incl_acc with "[//] Htok") as (q') "[Htok Hclose]". done.
  iInv in_atN as (V0) ">Hown" "Hclose'". simpl.
  iDestruct (monPred_in_intro with "Htok") as (V) "[#HV Htok]".
  iMod (idx_bor_acc_internal with "LFT Hs Hown Htok") as (Vb) "(% & HP & Hclose'')".
    solve_ndisj. solve_ndisj.
  iMod fupd_intro_mask' as "Hclose'''"; last iModIntro. solve_ndisj.
  iExists Vb. iSplitL "HP".
  { iIntros "HVb". iApply "HPP'". iApply (monPred_in_elim with "HVb").
    by rewrite monPred_at_later. }
  iIntros "HP". iMod "Hclose'''" as "_".
  iAssert ((monPred_in Vb → ▷ P') ∗ monPred_in V)%I with "[HP]" as "HH".
  { iIntros "{$HV} ?". iApply "HPP'". by iApply "HP". }
  iDestruct (monPred_in_intro with "HH") as (V') "(#HV' & HP & %)".
  iMod ("Hclose''" $! V' with "[//] [HP]") as "[Hown Htok]";
    [by rewrite -monPred_at_later; iApply "HP"; solve_lat|].
  iMod ("Hclose'" with "[Hown]") as "_"; [by iExists _; iFrame|].
  iApply "Hclose". by iApply (monPred_in_elim _ V').
Qed.

Lemma in_at_bor_shorten κ κ' P : (κ ⊑ κ') -∗ &in_at{κ'}P -∗ &in_at{κ}P.
Proof.
  iIntros "#H⊑ H". iDestruct "H" as (P' i) "(? & ?)".
  iExists P', i. iFrame "%∗". by iApply lft_incl_trans.
Qed.
End in_at.
