From lrust.lifetime Require Import borrow accessors.
From iris.algebra Require Import csum auth excl frac gmap agree gset.
From iris.proofmode Require Import tactics.
Set Default Proof Using "Type*".

Lemma and_extract_ownM {M} σ (P : uPred M) :
  uPred_ownM σ ∧ P -∗ uPred_ownM σ ∗ (uPred_ownM σ -∗ P).
Proof.
  uPred.unseal. split=>?? Hv[[σ' Hσ']]. revert Hv. move:Hσ'=>->. exists σ, σ'.
  split; [done|split].
  - exists ε. by rewrite right_id.
  - intros ????[σ'' ->]. eapply uPred_mono; [done| |done].
    rewrite assoc (comm op σ' σ). by exists σ''.
Qed.

Lemma and_extract_own `{inG Σ A} γ (x : A) (P : iProp Σ) :
  own γ x ∧ P -∗ own γ x ∗ (own γ x -∗ P).
Proof. rewrite own_eq. apply and_extract_ownM. Qed.

Lemma and_extract_own_bor `{lftG Lat E0 Σ} κ σ (P : iProp Σ) :
  own_bor κ σ ∧ P -∗ own_bor κ σ ∗ (own_bor κ σ -∗ P).
Proof.
  rewrite /own_bor bi.and_exist_r bi.exist_elim=>// γ.
  rewrite bi.sep_and -(assoc bi_and) (and_extract_own _ _ P).
  iIntros "(#Hag & Hσ & Hclose)". iSplitL "Hσ"; [by auto|]. iIntros "H".
  iApply "Hclose". iDestruct "H" as (γ') "[#Hag' H]".
  iDestruct (own_valid_2 with "Hag Hag'") as %Hγs. move : Hγs.
  rewrite auth_frag_valid op_singleton singleton_valid=> /agree_op_invL' <- //.
Qed.

Section reborrow.
Context `{!invG Σ, !lftG Lat E0 Σ, LatBottom Lat}.
Implicit Types κ : lft.

(* Notice that taking lft_inv for both κ and κ' already implies
   κ ≠ κ'.  Still, it is probably more instructing to require
   κ ⊂ κ' in this lemma (as opposed to just κ ⊆ κ'), and it
   should not increase the burden on the user. *)
Lemma raw_bor_unnest Vκ' V V' E A I Pb Pi P κ i Q κ' :
  ↑borN ⊆ E → V' ⊑ Vκ' → V' ⊑ V →
  let Iinv := (own_ilft_auth I ∗ ▷ lft_inv A κ)%I in
  κ ⊂ κ' →
  lft_alive_in A κ' →
  ▷ □ (Q ↔ idx_bor_own (κ, i)) V -∗
  Iinv -∗ ▷ Q V' -∗ slice borN i P -∗ ▷ lft_bor_alive κ' Pb Vκ' -∗
  ▷ lft_vs κ' (Q ∗ Pb) Pi Vκ' ={E}=∗ ∃ Pb',
    Iinv ∗ raw_bor κ' P V ∗ ▷ lft_bor_alive κ' Pb' Vκ' ∗ ▷ lft_vs κ' Pb' Pi Vκ'.
Proof.
  iIntros (? HV'κ' ? Iinv Hκκ' Haliveκ') "#HQ (HI & Hκ) HQV' #Hislice Hκalive' Hvs".
  rewrite lft_vs_unfold. iDestruct "Hvs" as (Vκ'' n) "(>% & >Hn● & Hvs)".
  iMod (own_cnt_update with "Hn●") as "[Hn● H◯]".
  { apply auth_update_alloc, (nat_local_update _ 0 (S n) 1); lia. }
  iAssert (idx_bor_own (κ, i) V ∧ ▷ Q V')%I with "[>HQV']" as "Hi".
  { iApply fupd_except_0. iModIntro. iSplit; [|done]. iApply timeless.
    iApply "HQ"; auto. }
  rewrite {2}/idx_bor_own monPred_at_exist /= bi.and_exist_r.
  iDestruct "Hi" as (V0) "Hi". rewrite monPred_at_sep monPred_at_embed monPred_at_in.
  rewrite (bi.sep_and ⌜V0 ⊑ V⌝%I) -(assoc bi_and) and_extract_own_bor.
  iDestruct "Hi" as "(% & Hi & HcloseQ)".
  iDestruct (own_bor_auth with "HI Hi") as %?.
  assert (κ ⊆ κ') by (by apply strict_include).
  iDestruct (lft_inv_alive_in with "Hκ") as (Vκ) "[>% Hκ]";
    first by eauto using lft_alive_in_subseteq.
  rewrite lft_inv_alive_unfold;
    iDestruct "Hκ" as (Pb' Pi') "(Hκalive & Hvs' & Hinh')".
  rewrite {2}/lft_bor_alive; iDestruct "Hκalive" as (B) "[[Hbox >HB●] HB]".
  iDestruct (own_bor_valid_2 with "HB● Hi")
    as %[HB%to_borUR_included _]%auth_both_valid.
  iMod (slice_empty _ _ true with "Hislice Hbox") as "[HP Hbox]"; first done.
  { by rewrite lookup_fmap HB. }
  iMod (own_bor_update_2 with "HB● Hi") as "[HB● Hi]".
  { eapply auth_update, singleton_local_update; last eapply
     (exclusive_local_update _ (Excl (Bor_rebor κ' V0))); last done.
    rewrite /to_borUR lookup_fmap. by rewrite HB. }
  iAssert (▷ lft_inv A κ)%I with "[H◯ HB● HB Hvs' Hinh' Hbox]" as "Hκ".
  { iNext. rewrite /lft_inv. iExists Vκ. iSplit; [done|]. iLeft.
    iSplit; last by eauto using lft_alive_in_subseteq.
    rewrite lft_inv_alive_unfold. iExists Pb', Pi'. iFrame "Hvs' Hinh'".
    rewrite /lft_bor_alive. iExists (<[i:=Bor_rebor κ' V0]>B).
    rewrite /to_borUR !fmap_insert. iFrame "Hbox HB●".
    rewrite big_sepM_delete //. simpl. iDestruct "HB" as "[% HB]".
    rewrite (big_sepM_delete _ (<[_:=_]>_) i); last by rewrite lookup_insert.
    rewrite -insert_delete delete_insert ?lookup_delete //=. iFrame. auto. }
  clear B HB Pb' Pi'.
  rewrite {1}/lft_bor_alive. iDestruct "Hκalive'" as (B) "[[Hbox >HB●] HB]".
  set (P' := ((monPred_in V ∨ ⎡P V0⎤) ∧ (monPred_in V0 → P))%I).
  iMod (slice_insert_full _ _ true _ _ P' bot with "[HP] Hbox")
    as (j) "(HBj & #Hjslice & Hbox)"; [done|iSplit; [|iIntros (???)]; auto|].
  iDestruct "HBj" as %HBj; move: HBj; rewrite lookup_fmap fmap_None=> HBj.
  iMod (own_bor_update with "HB●") as "[HB● Hj]".
  { apply auth_update_alloc,
     (alloc_singleton_local_update _ j (Excl (Bor_in bot))); last done.
    rewrite /to_borUR lookup_fmap. by rewrite HBj. }
  iModIntro. iExists (P' ∗ Pb)%I. rewrite /Iinv. iFrame "HI Hκ". iSplitL "Hj".
  { rewrite /raw_bor /idx_bor_own. iExists j.
    iSplitL; iExists _; iFrame "∗#"; [by auto|]. iIntros "!>!>". iSplit.
    - iIntros (??) "[_ H]". iApply "H". solve_lat.
    - iIntros (??) "H". iSplit; [|iIntros (???)]; by auto. }
  iSplitL "Hbox HB● HB".
  { rewrite /lft_bor_alive. iNext. iExists (<[j:=(Bor_in bot)]> B).
    rewrite /to_borUR !fmap_insert big_sepM_insert //=. auto with iFrame. }
  clear dependent Iinv I.
  iNext. rewrite lft_vs_unfold. iExists Vκ', (S n). iFrame "Hn●". iSplit; [done|].
  iIntros (I Vs) "% % Hinv [HP HPb] #Hκ†".
  rewrite {1}lft_vs_inv_unfold; iDestruct "Hinv" as "(Hκ'dead & HI & Hinv)".
  iDestruct (own_bor_auth with "HI Hi") as %?.
  iAssert ⌜κ' ∈ dom (gset _) I⌝%I as "%" .
  { rewrite /lft_bor_dead. iDestruct "Hκ'dead" as (??) "[H _]".
    iApply (own_bor_auth with "HI H"). }
  iDestruct (big_sepS_delete _ _ κ with "Hinv") as "[Hκalive Hinv]"; [done|].
  rewrite lft_inv_alive_unfold.
  iDestruct ("Hκalive" with "[//]") as (Pb' Pi') "(Hκalive & Hvs' & Hinh)".
  rewrite /lft_bor_alive; iDestruct "Hκalive" as (B') "[[Hbox HB●] HB]".
  iDestruct (own_bor_valid_2 with "HB● Hi")
    as %[HB%to_borUR_included _]%auth_both_valid.
  iAssert ⌜V0 ⊑ Vs κ⌝%I with "[#]" as "%".
  { rewrite (big_sepM_lookup _ _ _ _ HB) /=. iDestruct "HB" as "[$  _]". }
  iAssert (∃ VP, ⌜VP ⊑ Vs κ⌝ ∗ ▷ Q (Vs κ') ∗ ▷ P VP ∗
                 own_bor κ (● to_borUR (<[i:=Bor_in VP]>B')))%I
    with "[> HP Hi HcloseQ HB●]" as (VP) "(% & Hi & HP & HB●)".
  { rewrite bi.and_or_r. iDestruct "HP" as "[[>% HP]|[HP _]]".
    - iExists (Vs κ'). iSplitR; [by auto|].
      iMod (own_bor_update_2 with "HB● Hi") as "[HB● Hi]".
      { eapply auth_update, singleton_local_update,
          (exclusive_local_update _ (Excl (Bor_in (Vs κ')))); last done.
        rewrite /to_borUR lookup_fmap. by rewrite HB. }
      rewrite /to_borUR fmap_insert. iFrame. iSplitR "HP".
      + iDestruct "HQ" as "[_ HQ]". iApply "HQ". rewrite /idx_bor_own.
        iExists _. by iFrame.
      + iModIntro. iApply "HP". solve_lat.
    - iExists V0. iFrame "∗%". rewrite /to_borUR fmap_insert.
      iMod (own_bor_update_2 with "HB● Hi") as "[$ Hi]".
      { eapply auth_update, singleton_local_update,
          (exclusive_local_update _ (Excl (Bor_in V0))); last done.
        rewrite /to_borUR lookup_fmap. by rewrite HB. }
      by iDestruct ("HcloseQ" with "Hi") as "$". }
  iApply fupd_trans. iApply (fupd_mask_mono (↑borN)); [set_solver+|].
  iMod (slice_fill _ _ false with "Hislice HP Hbox") as "Hbox"; first by solve_ndisj.
  { by rewrite lookup_fmap HB. }
  iModIntro.
  rewrite big_sepM_delete //. simpl. iDestruct "HB" as "[(% & _ & H1◯) HB]".
  iMod ("Hvs" $! I Vs with "[%] [//] [Hκ'dead HI Hinv Hvs' Hinh HB● Hbox HB]
                         [$HPb $Hi] Hκ†") as "($ & $ & Hcnt')"; [solve_lat| |].
  { rewrite lft_vs_inv_unfold. iFrame "Hκ'dead HI".
    iApply (big_sepS_delete _ (dom (gset lft) I) with "[- $Hinv]"); [done|].
    iIntros (_). rewrite lft_inv_alive_unfold.
    iExists Pb', Pi'. iFrame "Hvs' Hinh". rewrite /lft_bor_alive.
    iExists (<[i:=Bor_in VP]>B'). rewrite /to_borUR !fmap_insert. iFrame.
    rewrite -insert_delete big_sepM_insert ?lookup_delete //=. iFrame. auto. }
  rewrite -[S n]Nat.add_1_l -nat_op_plus auth_frag_op own_cnt_op. by iFrame.
Qed.

Lemma raw_idx_bor_unnest E κ κ' i P :
  ↑lftN ⊆ E → κ ⊂ κ' →
  ⎡lft_ctx⎤ -∗ ⎡slice borN i P⎤ -∗
   raw_bor κ' (idx_bor_own (κ, i)) ={E}=∗ raw_bor κ' P.
Proof.
  iIntros (? Hκκ') "#LFT #Hs Hraw".
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  iDestruct (raw_bor_inI with "HI Hraw") as %HI'.
  rewrite (big_sepS_delete _ _ κ') // [_ A κ']/lft_inv. iDestruct "Hinv" as "[H Hinv]".
  iDestruct "H" as (Vκ) "[>% [[Hinvκ >%]|[Hinvκ >%]]]"; last first.
  { rewrite /lft_inv_dead. iDestruct "Hinvκ" as (Pi V0) "(>% & Hbordead & H)".
    iMod (raw_bor_fake with "[Hbordead]") as "[Hbordead $]"; [solve_ndisj|by iNext|].
    iMod ("Hclose" with "[-]"); [|done]. iExists _, _. iFrame.
    rewrite (big_sepS_delete _ (dom _ _) κ') //. iFrame.
    rewrite /lft_inv /lft_inv_dead. iExists _. iFrame "%". iRight. iFrame "%".
    iExists _, _. iFrame "∗%". by iNext. }
  rewrite {1}/raw_bor. iDestruct "Hraw" as (i') "[Hbor H]".
  iDestruct "H" as (P') "[#HP' #Hs']".
  rewrite lft_inv_alive_unfold /lft_bor_alive.
  iDestruct "Hinvκ" as (Pb Pi) "(Halive & Hvs & Hinh)".
  iDestruct "Halive" as (B) "[[Hbox >H●] HB]". rewrite [_ (κ', i')]/idx_bor_own.
  iDestruct "Hbor" as (V') "[#HV' Hbor]".
  iDestruct (own_bor_valid_2 with "H● Hbor")
    as %[EQB%to_borUR_included _]%auth_both_valid.
  iMod (slice_empty _ _ true with "Hs' Hbox") as "[Hidx Hbox]".
    solve_ndisj. by rewrite lookup_fmap EQB.
  iAssert (▷ ⌜κ ∈ dom (gset lft) I⌝)%I with "[#]" as ">%".
  { iNext. iDestruct (monPred_in_elim with "HV' Hidx") as "Hidx".
    iDestruct "HP'" as "[HP' _]". rewrite /idx_bor_own.
    iDestruct ("HP'" with "Hidx") as (?) "[_ Hown]".
    rewrite -embed_pure. iApply (own_bor_auth with "HI Hown"). }
  rewrite (big_sepS_elem_of_acc _ _ κ); last first.
  { by rewrite elem_of_difference not_elem_of_singleton; eauto using strict_ne. }
  iDestruct "Hinv" as "[Hinvκ Hclose']".
  iDestruct (slice_delete_empty with "Hs' Hbox") as (Pb') "[#HeqPb' Hbox]";
    [by apply lookup_insert|].
  iMod (own_bor_update_2 with "H● Hbor") as "H●".
  { apply auth_update_dealloc, delete_singleton_local_update. apply _. }
  iAssert (▷ ⌜V' ⊑ Vκ⌝)%I with "[#]" as ">%".
  { iNext. rewrite big_sepM_lookup //. simpl. by iDestruct "HB" as %?. }
  iCombine "HP' HV'" as "HP'V'".
  iDestruct (monPred_in_intro with "HP'V'") as (V) "(HV & HP'V & %)".
  iMod (raw_bor_unnest Vκ with "[HP'V] [$HI $Hinvκ] Hidx Hs [Hbox H● HB] [Hvs]")
    as (Pb'') "HH"; [solve_ndisj|done|done|done|done| | | |].
  { iIntros "!>!>!>". by iDestruct  "HP'V" as "#HP'V". }
  { rewrite /lft_bor_alive (big_sepM_delete _ B i') //. iDestruct "HB" as "[_ HB]".
    iExists (delete i' B). rewrite -fmap_delete. iFrame.
    rewrite fmap_delete -insert_delete delete_insert ?lookup_delete //=. }
  { iModIntro. iModIntro. iApply (lft_vs_cons with "[] Hvs").
    iIntros (??) "HD %% [??] %% _ !>". iFrame "HD". iNext. iRewrite "HeqPb'". iFrame. }
  iDestruct "HH" as "([HI Hinvκ] & Hb & Halive & Hvs)".
  iMod ("Hclose" with "[-Hb]"); last first.
  { iApply monPred_in_elim. done. by iFrame. }
  iExists _, _. iFrame. rewrite (big_sepS_delete _ (dom _ _) κ') //.
  iDestruct ("Hclose'" with "Hinvκ") as "$".
  rewrite /lft_inv. iExists Vκ. iSplit; [done|].
  rewrite lft_inv_alive_unfold. iLeft. iSplit; [|done]. iExists _, _. iFrame.
Qed.

Lemma raw_bor_shorten E κ κ' P :
  ↑lftN ⊆ E → κ ⊆ κ' →
  ⎡lft_ctx⎤ -∗ raw_bor κ P ={E}=∗ raw_bor κ' P.
Proof.
  iIntros (? Hκκ') "#LFT Hbor".
  destruct (decide (κ = κ')) as [<-|Hκneq]; [by iFrame|].
  assert (κ ⊂ κ') by (by apply strict_spec_alt).
  rewrite [_ κ P]/raw_bor. iDestruct "Hbor" as (s) "[Hbor Hs]".
  iDestruct "Hs" as (P') "[#HP'P #Hs]".
  iMod (raw_bor_create _ κ' with "LFT Hbor") as "[Hbor _]"; [done|].
  iApply (raw_bor_iff with "HP'P"). by iApply raw_idx_bor_unnest.
Qed.

Lemma idx_bor_unnest E κ κ' i P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ,i}P -∗ &{κ'}(idx_bor_own i) ={E}=∗ &{κ ⊓ κ'}P.
Proof.
  iIntros (?) "#LFT #HP Hbor".
  rewrite [(&{κ'}_)%I]/bor. iDestruct "Hbor" as (κ'0) "[#Hκ'κ'0 Hbor]".
  destruct (decide (κ'0 = static)) as [->|Hκ'0].
  { iMod (bor_acc_strong with "LFT [Hbor] []") as (?) "(_ & Hbor & _)";
      [done| |iApply (lft_tok_static 1)|].
    - rewrite /bor. iExists static. iFrame. iApply lft_incl_refl.
    - iDestruct "Hbor" as ">Hbor". rewrite bor_unfold_idx /=.
      iExists i. iFrame. iApply idx_bor_shorten; [|done].
      iApply lft_intersect_incl_l. }
  rewrite /idx_bor /bor. destruct i as [κ0 i].
  iDestruct "HP" as "[Hκκ0 HP]". iDestruct "HP" as (P') "[HPP' HP']".
  iMod (raw_bor_shorten _ _ (κ0 ⊓ κ'0) with "LFT Hbor") as "Hbor";
    [done|by apply gmultiset_disj_union_subseteq_r|].
  iMod (raw_idx_bor_unnest with "LFT HP' Hbor") as "Hbor";
    [done|by apply gmultiset_disj_union_subset_l|].
  iExists _. iDestruct (raw_bor_iff with "HPP' Hbor") as "$".
    by iApply lft_intersect_mono.
Qed.
End reborrow.
