From iris.proofmode Require Import tactics.
From lrust.lang Require Import notation.
From lrust.lifetime Require Import at_borrow.
From gpfsl.gps Require Import middleware.

Section gps_at_bor_SP.
Context `{!noprolG Σ, !gpsG Σ Prtcl, !lftG view_Lat (↑histN) Σ} (N: namespace).
Local Notation vProp := (vProp Σ).

Implicit Types (IP : interpO Σ Prtcl) (IPC: interpCasO Σ Prtcl) (l : loc)
               (s : pr_stateT Prtcl) (t : time) (v : val) (E : coPset) (q: Qp)
               (κ : lft) (γ : gname).

Definition GPS_aSP_Reader IP IPC l κ t s v γ : vProp :=
  (GPS_SWReader l t s v γ ∗ &at{κ, N} (GPS_INV IP l IPC γ))%I.
Definition GPS_aSP_Writer IP IPC l κ t s v γ : vProp :=
  (GPS_SWWriter l t s v γ ∗ &at{κ, N} (GPS_INV IP l IPC γ))%I.
Definition GPS_aSP_SharedReader IP IPC l κ t s v q γ : vProp :=
  (GPS_SWSharedReader l t s v q γ ∗ &at{κ, N} (GPS_INV IP l IPC γ))%I.
Definition GPS_aSP_SharedWriter IP IPC l κ t s v γ : vProp :=
  (GPS_SWSharedWriter l t s v γ ∗ &at{κ, N} (GPS_INV IP l IPC γ))%I.

Lemma GPS_aSP_Init IP IPC l κ s q P (Q: time → val → gname → vProp) E
  (DISJ: N ## lftN) (SUB: ↑lftN ⊆ E):
  ⎡ lft_ctx ⎤ -∗ (q).[κ] -∗ ▷ ⎡ hist_inv ⎤ -∗
  &{κ} (∃ v, l ↦ v ∗ P v) -∗
  (∀ t γ v, ▷ P v -∗ GPS_SWWriter l t s v γ ={E}=∗ ▷ IP true l γ t s v ∗ Q t v γ) -∗
  (□ ∀ t γ s v, ▷ IP true l γ t s v ={↑histN}=∗ ▷ P v) ={E}=∗
  (q).[κ] ∗ ∃ γ t v, GPS_aSP_Reader IP IPC l κ t s v γ ∗ Q t v γ.
Proof.
  iIntros "#LFT Htok #HINV Hl HP #HP2".
  iMod (bor_acc_cons with "LFT Hl Htok") as "[Hl Hclose]"; first done.
  iDestruct "Hl" as (v) "[>Hl P]".
  set Q' : time → gname → vProp :=
    (λ t γ, Q t v γ ∗ GPS_SWReader l t s v γ)%I.
  iMod (GPS_SWRaw_Init_vs_strong IP l IPC true _ s Q' with "Hl [HP P]")
    as (γ t) "[Inv [Q R]]".
  { iIntros (t γ) "W". iDestruct (GPS_SWWriter_Reader with "W") as "#$".
    by iApply ("HP" with "P W"). }
  iMod ("Hclose" with "[] Inv") as "[Hl Htok]".
  { iIntros "!> Inv".
    iMod (GPS_INV_dealloc IP l IPC true with "HINV Inv")
      as (t' s' v') "[Hl IP]"; [done|].
    iMod ("HP2" with "IP") as "P'".
    iIntros "!> !>". iExists _. by iFrame. }
  iMod (bor_share N with "Hl") as "Hl"; [done|].
  iFrame "Htok". iExists γ, t, v. by iFrame "∗".
Qed.

Lemma GPS_aSP_Read IP IPC l κ q R (o: memOrder) t s v γ tid E
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (RLX: o = Relaxed ∨ o = AcqRel) :
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aSP_Reader IP IPC l κ t s v γ
      ∗ (▷ ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
           <obj> ((IP false l γ t' s' v' ={E ∖ ↑N}=∗
                      IP false l γ t' s' v' ∗ R t' s' v') ∧
                  (IP true l γ t' s' v' ={E ∖ ↑N}=∗
                      IP true l γ t' s' v' ∗ R t' s' v') ∧
                  (IPC l γ t' s' v' ={E ∖ ↑N}=∗
                      IPC l γ t' s' v' ∗ R t' s' v'))) }}}
    Read o #l in tid @ E
  {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝
      ∗ GPS_aSP_Reader IP IPC l κ t' s' v' γ ∗ (q).[κ]
      ∗ if decide (AcqRel ⊑ o)%stdpp then R t' s' v' else ▽{tid} R t' s' v' }}}.
Proof.
  iIntros (Φ) "(#LFT & Htok & #[Hlc Hshr] & VS) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iApply (GPS_SWRaw_Read IP l IPC R o t s v γ tid Vb
          with "[$Hlc $Hproto $VS]"); [solve_ndisj|done|..].
  iNext. iIntros (t' s' v') "(Ext & Hlc' & Hproto & R)".
  iMod ("Hclose" with "Hproto") as "Htok".
  iApply ("Post" $! t' s' v' with "[$Ext $Hlc' $Hshr $Htok $R]").
Qed.

Lemma GPS_aSP_SWWrite IP IPC l κ q R o t s s' v v' γ tid E
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (RLX: o = Relaxed ∨ o = AcqRel) (Exts: s ⊑ s') :
  let WVS: vProp :=
  (∀ t', ⌜(t < t')%positive⌝ ={E ∖ ↑N}=∗ IP true l γ t' s' v')%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aSP_Writer IP IPC l κ t s v γ
      ∗ ▷ (if decide (AcqRel ⊑ o)%stdpp then WVS else △{tid} WVS)
      ∗ ▷ <obj> (IP true l γ t s v ={E ∖ ↑N}=∗ IPC l γ t s v ∗ R) }}}
    Write o #l v' in tid @ E
  {{{ t', RET #☠; ⌜(t < t')%positive⌝ ∗ GPS_aSP_Writer IP IPC l κ t' s' v' γ
      ∗ (q).[κ] ∗ R }}}.
Proof.
  iIntros (VS Φ) "(#LFT & Htok & [W #Hshr] & VS) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iApply (GPS_SWRaw_SWWrite IP l IPC R o t s s' v v' γ tid Vb
          with "[$W $Hproto $VS]"); [done|solve_ndisj|done|..].
  iNext. iIntros (t') "(Ext & W & Hproto & R)".
  iMod ("Hclose" with "Hproto") as "Htok".
  iApply ("Post" $! t' with "[$Ext $W $Hshr $Htok $R]").
Qed.

Lemma GPS_aSP_SWWrite_rel IP IPC l κ q Q Q1 Q2 t s s' v v' γ tid E
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (Exts: s ⊑ s') :
  let WVS: vProp :=
  (∀ t', ⌜(t < t')%positive⌝ -∗ GPS_SWWriter l t' s' v' γ -∗ Q2
              ={E ∖ ↑N}=∗ (<obj> (Q1 ={E ∖ ↑N}=∗ IPC l γ t s v)) ∗
                     |={E ∖ ↑N}=> (IP true l γ t' s' v' ∗ Q t'))%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aSP_Writer IP IPC l κ t s v γ
      ∗ ▷ WVS ∗ ▷ <obj> (IP true l γ t s v ={E ∖ ↑N}=∗ Q1 ∗ Q2) }}}
    #l <-ʳᵉˡ v' in tid @ E
  {{{ t', RET #☠; ⌜(t < t')%positive⌝ ∗ GPS_aSP_Reader IP IPC l κ t' s' v' γ
      ∗ (q).[κ] ∗ Q t' }}}.
Proof.
  iIntros (VS Φ) "(#LFT & Htok & [W #Hshr] & VS) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iApply (GPS_SWRaw_SWWrite_rel IP l IPC Q Q1 Q2 t s s' v v' γ tid Vb
          with "[$W $Hproto $VS]"); [done|solve_ndisj|..].
  iNext. iIntros (t') "(Ext & R' & Hproto & Q)".
  iMod ("Hclose" with "Hproto") as "Htok".
  iApply ("Post" $! t' with "[$Ext $R' $Hshr $Htok $Q]").
Qed.

Lemma GPS_aSP_SharedWriter_CAS_int_weak
  IP IPC l κ q orf or ow (vr: Z) (vw: val) t s v γ P Q Q1 Q2 R E E' tid
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (SUBE: E' ⊆ E ∖ ↑N) (RLXF: orf = Relaxed ∨ orf = AcqRel)
  (RLXR: or = Relaxed ∨ or = AcqRel) (RLXW: ow = Relaxed ∨ ow = AcqRel) :
  let VSC : vProp :=
    (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            (IP true l γ t' s' v' ∨ IP false l γ t' s' v' ∨ IPC l γ t' s' v') -∗
            ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
  let VS : vProp :=
    (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
        ((▷ (P -∗ IPC l γ t' s' #vr ={E ∖ ↑N}=∗ False)) ∧
         ((<obj> (▷ IP true l γ t' s' #vr ={E ∖ ↑N}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
         (P -∗ ▷ Q2 t' s' ={E ∖ ↑N}=∗ ▷ GPS_SWSharedWriter l t' s' #vr γ ∗
            ∃ s'', ⌜s' ⊑ s''⌝ ∗
              (∀ t'' , ⌜(t' < t'')%positive⌝ -∗
                  ▷ GPS_SWSharedWriter l t'' s'' vw γ ={E ∖ ↑N}=∗
                  (<obj> (▷ Q1 t' s' ={E ∖ ↑N}=∗ ▷ IP false l γ t' s' #vr)) ∗
                  |={E ∖ ↑N,E'}▷=> Q t'' s'' ∗ ▷ IP true l γ t'' s'' vw))) ) ∧
        ▷ (∀ (v: lit), ⌜lit_neq vr v⌝ -∗
        <obj> ((IP false l γ t' s' #v ={E ∖ ↑N}=∗ IP false l γ t' s' #v ∗ R t' s' v) ∧
               (IP true l γ t' s' #v ={E ∖ ↑N}=∗ IP true l γ t' s' #v ∗ R t' s' v) ∧
               (IPC l γ t' s' #v ={E ∖ ↑N}=∗ IPC l γ t' s' #v ∗ R t' s' v))))%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aSP_Reader IP IPC l κ t s v γ
      ∗ (▷ VSC)
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then VS else △{tid} VS)
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then P  else △{tid} P) }}}
    CAS #l #vr vw orf or ow in tid @ E
  {{{ (b: bool) t' s' (v': lit), RET #b;
      (q).[κ] ∗ ⌜s ⊑ s'⌝
      ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
            ∗ GPS_aSP_Reader IP IPC l κ t' s' vw γ
            ∗ if decide (AcqRel ⊑ or)%stdpp then Q t' s' else ▽{tid} Q t' s')
        ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
            ∗ GPS_aSP_Reader IP IPC l κ t' s' #v' γ
            ∗ (if decide (AcqRel ⊑ orf)%stdpp
               then R t' s' v' else ▽{tid} (R t' s' v'))
            ∗ (if decide (AcqRel ⊑ ow)%stdpp then P else △{tid} P) ) ) }}}.
Proof.
  iIntros (VSC VS Φ) "(#LFT & Htok & #[Hlc Hshr] & VSC & VS & P) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iApply (GPS_SWRaw_SharedWriter_CAS_int_weak IP l IPC orf or ow vr vw t s v γ
            P Q Q1 Q2 R tid Vb with "[$Hlc $Hproto $VSC $VS $P]");
    [done|solve_ndisj|done|done|done|..].
  iIntros "!>" (b t' s' v') "(Hproto & Ext & CASE)".
  iMod ("Hclose" with "Hproto") as "Htok".
  iApply ("Post" $! b t' s' v' with "[-$Htok $Ext]").
  iDestruct "CASE" as "[?|?]"; [iLeft|iRight]; iFrame "Hshr ∗".
Qed.

Lemma GPS_aSP_SharedWriter_CAS_int_strong
  IP IPC l κ q qs (dropR: bool) orf or ow (vr: Z) (vw: val) t s v γ
  P Q Q1 Q2 R E E' tid
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (SUBE: E' ⊆ E ∖ ↑N) (RLXF: orf = Relaxed ∨ orf = AcqRel)
  (RLXR: or = Relaxed ∨ or = AcqRel) (RLXW: ow = Relaxed ∨ ow = AcqRel) :
  let VSC : vProp :=
    (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            (IP true l γ t' s' v' ∨ IP false l γ t' s' v') -∗
            ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
  let VS : vProp :=
    (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
      ((<obj> (▷ IP true l γ t' s' #vr ={E ∖ ↑N}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
       (P -∗ ▷ Q2 t' s' ={E ∖ ↑N}=∗ ▷ GPS_SWSharedWriter l t' s' #vr γ ∗
          ∃ s'', ⌜s' ⊑ s''⌝ ∗
            (∀ t'' , ⌜(t' < t'')%positive⌝ -∗ ▷ GPS_SWSharedWriter l t'' s'' vw γ -∗
              (if dropR then ▷ GPS_SWSharedReader l t'' s'' vw qs γ else True) ={E ∖ ↑N}=∗
              (<obj> (▷ Q1 t' s' ={E ∖ ↑N}=∗ ▷ IP false l γ t' s' #vr)) ∗
              |={E ∖ ↑N,E'}▷=> Q t'' s'' ∗ ▷ IP true l γ t'' s'' vw)) ) ∧
      ▷ (∀ (v: lit), ⌜lit_neq vr v⌝ -∗
        <obj> ((IP false l γ t' s' #v ={E ∖ ↑N}=∗ IP false l γ t' s' #v ∗ R t' s' v) ∧
               (IP true l γ t' s' #v ={E ∖ ↑N}=∗ IP true l γ t' s' #v ∗ R t' s' v))))%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aSP_SharedReader IP IPC l κ t s v qs γ
      ∗ (▷ VSC)
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then VS else △{tid} VS)
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then P  else △{tid} P) }}}
    CAS #l #vr vw orf or ow in tid @ E
  {{{ (b: bool) t' s' (v': lit), RET #b;
      (q).[κ] ∗ ⌜s ⊑ s'⌝
      ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
            (* TODO: GPS_aSP_Reader here *)
            ∗ (if dropR then True else GPS_aSP_SharedReader IP IPC l κ t' s' vw qs γ)
            ∗ if decide (AcqRel ⊑ or)%stdpp then Q t' s' else ▽{tid} Q t' s')
        ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
            ∗ GPS_aSP_SharedReader IP IPC l κ t' s' #v' qs γ
            ∗ (if decide (AcqRel ⊑ orf)%stdpp then R t' s' v' else ▽{tid} (R t' s' v'))
            ∗ (if decide (AcqRel ⊑ ow)%stdpp then P else △{tid} P) ) ) }}}.
Proof.
  iIntros (VSC VS Φ) "(#LFT & Htok & [Hlc #Hshr] & VSC & VS & P) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iApply (GPS_SWRaw_SharedWriter_CAS_int_strong IP l IPC dropR orf or ow vr vw
            t s v qs γ P Q Q1 Q2 R tid Vb with "[$Hlc $Hproto $VS $VSC $P]");
    [done|solve_ndisj|done|done|done|..].
  iIntros "!>" (b t' s' v') "(Hproto & Ext & CASE)".
  iMod ("Hclose" with "Hproto") as "Htok".
  iApply ("Post" $! b t' s' v' with "[-$Htok $Ext]").
  iDestruct "CASE" as "[CASE|?]"; [iLeft|iRight]; last iFrame "Hshr ∗".
  iDestruct "CASE" as "($ & R & $)". destruct dropR; [done|]. by iFrame "R Hshr".
Qed.

Lemma GPS_aSP_SharedWriter_revert IP IPC l κ q t t' s s' v v' γ E
  (SUB1: ↑lftN ⊆ E) (SUB2: ↑N ⊆ E):
  ⎡ lft_ctx ⎤ -∗ (q).[κ] -∗
  GPS_aSP_SharedWriter IP IPC l κ t s v γ -∗
  GPS_SWSharedReader l t' s' v' 1%Qp γ ={E}=∗
  (q).[κ] ∗ GPS_aSP_Writer IP IPC l κ t s v γ.
Proof.
  iIntros "#LFT Htok [W #Hshr] SR".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose]"; [done..|].
  iMod (GPS_SWRaw_SharedWriter_revert with "W SR Hproto") as "[$ Hproto]".
  iMod ("Hclose" with "Hproto") as "$". by iFrame "Hshr".
Qed.

End gps_at_bor_SP.

Section gps_at_bor_PP.
Context `{!noprolG Σ, !gpsG Σ Prtcl, !gpsExAgG Σ, !lftG view_Lat (↑histN) Σ}
        (N: namespace).
Local Notation vProp := (vProp Σ).

Implicit Types (IP : interpO Σ Prtcl) (l : loc)
               (s : pr_stateT Prtcl) (t : time) (v : val) (E : coPset) (q: Qp)
               (κ : lft) (γ : gname).

Definition GPS_aPP IP l κ t s v γ : vProp :=
  (GPS_PP_Local l t s v γ ∗ &at{κ, N} (GPS_PPInv IP l γ))%I.

Global Instance GPS_aPP_ne l κ t s v γ: NonExpansive (λ IP, GPS_aPP IP l κ t s v γ).
Proof. move => ????. apply bi.sep_ne; [done|]. by apply at_bor_ne, GPS_PPInv_ne. Qed.
Global Instance GPS_aPP_ne_persistent: Persistent (GPS_aPP IP l κ t s v γ) := _.

Lemma GPS_aPP_lft_mono IP l κ κ' t s v γ :
  κ' ⊑ κ -∗ GPS_aPP IP l κ t s v γ -∗ GPS_aPP IP l κ' t s v γ.
Proof. iIntros "#Hincl". iDestruct 1 as "[$ Inv]". by iApply at_bor_shorten. Qed.

Lemma GPS_aPP_Init IP l κ s q P E (DISJ: N ## lftN) (SUB: ↑lftN ⊆ E):
  ⎡ lft_ctx ⎤ -∗ (q).[κ] -∗ ▷ ⎡ hist_inv ⎤ -∗
  &{κ} (∃ v, l ↦ v ∗ P v) -∗
  (∀ t γ v, ▷ P v -∗ ▷ IP true l γ t s v) -∗
  (□ ∀ t γ s v, ▷ IP true l γ t s v ={↑histN}=∗ ▷ P v) ={E}=∗
  (q).[κ] ∗ ∃ γ t v, GPS_aPP IP l κ t s v γ.
Proof.
  iIntros "#LFT Htok #HINV Hl HP #HP2".
  iMod (bor_acc_cons with "LFT Hl Htok") as "[Hl Hclose]"; first done.
  iDestruct "Hl" as (v) "[>Hl P]".
  iMod (GPS_PPRaw_Init_vs IP l true with "Hl [HP P]") as (γ t) "[PP Inv]".
  { iIntros (t γ). by iApply "HP". }
  iMod ("Hclose" with "[] Inv") as "[Hl Htok]".
  { iIntros "!> Inv".
    iMod (GPS_PPInv_dealloc IP l true with "HINV Inv")
      as (t' s' v') "[Hl IP]"; [done|].
    iMod ("HP2" with "IP") as "P'".
    iIntros "!> !>". iExists _. by iFrame. }
  iMod (bor_share N with "Hl") as "Hl"; [done|].
  iFrame "Htok". iExists γ, t, v. by iFrame "PP Hl".
Qed.

Lemma GPS_aPP_Read IP l κ q R (o: memOrder) t s v γ tid E
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (RLX: o = Relaxed ∨ o = AcqRel) :
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aPP IP l κ t s v γ
      ∗ (▷ ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
             <obj> ((IP false l γ t' s' v' ={E ∖ ↑N}=∗
                        IP false l γ t' s' v' ∗ R t' s' v') ∧
                    (IP true l γ t' s' v' ={E ∖ ↑N}=∗
                        IP true l γ t' s' v' ∗ R t' s' v'))) }}}
    Read o #l in tid @ E
  {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝ ∗ GPS_aPP IP l κ t' s' v' γ ∗ (q).[κ]
      ∗ if decide (AcqRel ⊑ o)%stdpp then R t' s' v' else ▽{tid} R t' s' v' }}}.
Proof.
  iIntros (Φ) "(#LFT & Htok & #[Hlc Hshr] & VS) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose1]"; [done..|].
  iApply (GPS_PPRaw_Local_Read IP l R o t s v γ tid Vb
          with "[$Hlc $Hproto $VS]"); [solve_ndisj|done|..].
  iNext. iIntros (t' s' v') "(Ext & Hlc' & Hproto & R)".
  iMod ("Hclose1" with "Hproto") as "Htok".
  iApply ("Post" $! t' s' v' with "[$Ext $Hlc' $Hshr $Htok $R]").
Qed.

Lemma GPS_aPP_Write IP l κ q (o: memOrder) t v v' s s' γ tid E
  (FIN: final_st s') (DISJ: histN ## N)
  (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E) (SUB3: ↑N ⊆ E)
  (RLX: o = Relaxed ∨ o = AcqRel) :
  let VS : vProp :=
    (∀ t' : time, ⌜(t < t')%positive⌝ -∗
          GPS_PP_Local l t' s' v' γ ={E ∖ ↑N}=∗ IP true l γ t' s' v')%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aPP IP l κ t s v γ
      ∗ ▷ if decide (AcqRel ⊑ o)%stdpp then VS else △{tid} VS }}}
    Write o #l v' in tid @ E
  {{{ t', RET #☠; GPS_aPP IP l κ t' s' v' γ ∗ (q).[κ] }}}.
Proof.
  iIntros (VS Φ) "(#LFT & Htok & #[Hlc Hshr] & VS) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose1]"; [done..|].
  iApply (GPS_PPRaw_Local_Write IP l o t v v' s s' γ Vb tid _ FIN
          with "[$Hlc $Hproto $VS]"); [solve_ndisj|done|..].
  iNext. iIntros (t') "[Hlc' Hproto]".
  iMod ("Hclose1" with "Hproto") as "Htok".
  iApply ("Post" $! t' with "[$Hlc' $Hshr $Htok]").
Qed.

Lemma GPS_aPP_CAS_int_simple IP l κ q t s v (vr: Z) (vw: val) orf or ow γ
  P Q Q1 Q2 R tid E
  (DISJ: histN ## N) (SUB1: ↑lftN ⊆ E) (SUB2: ↑histN ⊆ E ∖ ↑N) (SUB3: ↑N ⊆ E)
  (ORF: orf = Relaxed ∨ orf = AcqRel)
  (OR: or = Relaxed ∨ or = AcqRel) (OW: ow = Relaxed ∨ ow = AcqRel):
  let VSC : vProp :=
    (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                  (IP true l γ t' s' v' ∨ IP false l γ t' s' v') -∗
                  ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
  let VS : vProp :=
    (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
      ((<obj> (▷ IP true l γ t' s' #vr ={E ∖ ↑N}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
       (P -∗ ▷ Q2 t' s' ={E ∖ ↑N}=∗ ∃ s'', ⌜s' ⊑ s''⌝ ∗
          (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ (GPS_PP_Local l t s'' vw γ) ={E ∖ ↑N}=∗
              (<obj> (▷ Q1 t' s' ={E ∖ ↑N}=∗ ▷ IP false l γ t' s' #vr)) ∗
              |={E ∖ ↑N}▷=> Q t s'' ∗ ▷ IP true l γ t s'' vw)) ) ∧
      (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
          <obj> ((IP false l γ t' s' #v ={E ∖ ↑N}=∗
                    IP false l γ t' s' #v ∗ R t' s' v) ∧
                 (IP true l γ t' s' #v ={E ∖ ↑N}=∗
                    IP true l γ t' s' #v ∗ R t' s' v))))%I in
  {{{ ⎡ lft_ctx ⎤ ∗ (q).[κ] ∗ GPS_aPP IP l κ t s v γ
      ∗ ▷ VSC
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then VS else △{tid} VS)
      ∗ (if decide (AcqRel ⊑ ow)%stdpp then P  else △{tid} P ) }}}
    CAS #l #vr vw orf or ow in tid @ E
  {{{ (b: bool) t' s' (v': lit), RET #b;
        (q).[κ] ∗ ⌜s ⊑ s'⌝
        ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
              ∗ GPS_aPP IP l κ t' s' vw γ
              ∗ if decide (AcqRel ⊑ or)%stdpp then Q t' s' else ▽{tid} Q t' s')
          ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
              ∗ GPS_aPP IP l κ t' s' #v' γ
              ∗ (if decide (AcqRel ⊑ orf)%stdpp
                 then R t' s' v' else ▽{tid} (R t' s' v'))
              ∗ (if decide (AcqRel ⊑ ow)%stdpp
                 then P  else △{tid} P ))) }}}.
Proof.
  iIntros (VSC VS Φ) "(#LFT & Htok & #[Hlc Hshr] & VSC & VS & P) Post".
  iMod (at_bor_acc with "LFT Hshr Htok") as (Vb) "[Hproto Hclose1]"; [done..|].
  iApply (GPS_PPRaw_Local_CAS_int_simple IP l orf or ow v vr vw t s P Q Q1 Q2 R
            γ tid Vb with "[$Hlc $Hproto $VSC $VS $P]"); [done..|].
  iIntros "!>" (b t' s' v') "(Hproto & Ext & CASE)".
  iMod ("Hclose1" with "Hproto") as "Htok".
  iApply ("Post" $! b t' s' v' with "[$Htok $Ext CASE]").
  iDestruct "CASE" as "[?|?]"; [iLeft|iRight]; by iFrame "Hshr".
Qed.

End gps_at_bor_PP.
