From Coq.QArith Require Import Qcanon.
From iris.bi Require Import fractional.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import auth csum frac agree.
From gpfsl.gps Require Import middleware protocols.
From lrust.logic Require Import gps.
From lrust.lifetime Require Import at_borrow.
From lrust.typing Require Import typing.
From lrust.typing.lib.rwlock Require Import rwlock rwlockreadguard.
Set Default Proof Using "Type".

Section rwlockreadguard_functions.
  Context `{!typeG Σ, !rwlockG Σ}.

  (* Turning a rwlockreadguard into a shared borrow. *)
  Definition rwlockreadguard_deref : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
      let: "r'" := !"x'" +ₗ #1 in
      letalloc: "r" <- "r'" in
      delete [ #1; "x"];; return: ["r"].

  Lemma rwlockreadguard_deref_type ty `{!TyWf ty} :
    typed_val rwlockreadguard_deref
      (fn(∀ '(α, β), ∅; &shr{α}(rwlockreadguard β ty)) → &shr{α} ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros ([α β] ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x').
    iIntros (tid) "#LFT #HE Hna HL Hk HT". simpl_subst.
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']". destruct x' as [[|lx'|]|]; try done.
    iDestruct "Hx'" as (l') "#[Hfrac Hshr]".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)"; [solve_typing..|].
    iMod (frac_bor_acc with "LFT Hfrac Hα") as (qlx') "[H↦ Hcloseα]". done.
    rewrite heap_mapsto_vec_singleton. wp_read. wp_op. wp_let.
    iMod ("Hcloseα" with "[$H↦]") as "Hα". iMod ("Hclose" with "Hα HL") as "HL".
    iDestruct (lctx_lft_incl_incl α β with "HL HE") as "#Hαβ"; [solve_typing..|].
    iApply (type_type _ _ _ [ x ◁ box (&shr{α}(rwlockreadguard β ty));
                              #(l' +ₗ 1) ◁ &shr{α}ty]
      with "[] LFT HE Hna HL Hk"); first last.
    { rewrite tctx_interp_cons tctx_interp_singleton tctx_hasty_val tctx_hasty_val' //.
      iFrame. iApply (ty_shr_mono with "[] Hshr"). iApply lft_incl_glb; first done.
      by iApply lft_incl_refl. }
    iApply (type_letalloc_1 (&shr{α}ty)); [solve_typing..|].
    iIntros (r). simpl_subst. iApply type_delete; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  (* Dropping a rwlockreadguard and releasing the corresponding lock. *)
  Definition rwlockreadguard_drop : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
    withcont: "loop":
      "loop" []
    cont: "loop" [] :=
      let: "n" := !ʳˡˣ"x'" in
      if: CAS "x'" "n" ("n" - #1) Relaxed Relaxed AcqRel then
        delete [ #1; "x"];;
        let: "r" := new [ #0] in return: ["r"]
      else "loop" [].

  Lemma rwlockreadguard_drop_type ty `{!TyWf ty} :
    typed_val rwlockreadguard_drop (fn(∀ α, ∅; rwlockreadguard α ty) → unit).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (α ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
    iApply (type_cont [] [ϝ ⊑ₗ []] (λ _, [x ◁ _; x' ◁ _]));
      [iIntros (loop)|iIntros "/= !#"; iIntros (loop arg); inv_vec arg];
      simpl_subst.
    { iApply type_jump; solve_typing. }
    iIntros (tid) "#LFT #HE Hna HL Hk HT".
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']".
    destruct x' as [[|lx'|]|]; try done. simpl.
    iDestruct "Hx'" as (ν q γs β) "(#Hx' & #Hαβ & Hν & H◯ & #H† & Hinv)".
    iDestruct "Hinv" as (γ tyO tyS) "(R & #EqO & #EqS & R')".
    iDestruct "R'" as (tR vR) "R'".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)"; [solve_typing..|].
    iMod (lft_incl_acc with "Hαβ Hα") as (qβ) "[Hβ Hcloseα]". done.
    wp_bind (!ʳˡˣ#lx')%E.
    iApply (GPS_aSP_Read rwlockN _ _ _ β qβ
            (λ _ _ v, ∃ st, ⌜v = #(Z_of_rwlock_st st)⌝)%I with "[$LFT $Hβ $R']");
      [solve_ndisj|done..|by left| |].
    { iIntros "!>" (t' [] v [_ Ext']) "!>". iSplit; last iSplit.
      - iDestruct 1 as (st Eqvs) "F".
        iIntros "!>". iSplitR ""; iExists st; [by iFrame "F"|done].
      - iDestruct 1 as (st Eqvs) "F".
        iIntros "!>". iSplitR ""; iExists st; [by iFrame "F"|done].
      - iIntros (Eqvs) "!>". iSplit; [done|]. by iExists (Some (inl ())). }
    iIntros "!>" (t' [] v') "(_ & #R' & Hβ & Int)".
    iDestruct (acq_exist with "Int") as (st2) "Int".
    iDestruct (acq_pure_elim with "Int") as %?. subst v'. iClear "Int". clear tR vR.

    wp_let. wp_op. wp_bind (CAS _ _ _ _ _ _).
    set P : vProp Σ := (rwown γs (reading_st q ν) ∗ (q).[ν])%I.
    set Q : time → () → vProp Σ := (λ _ _, True)%I.
    set Q1: time → () → vProp Σ :=
      (λ t _, rwlock_interp γs β tyO tyS false lx' γ t () #(Z_of_rwlock_st st2))%I.
    set Q2: time → () → vProp Σ :=
      (λ t _, rwlock_interp γs β tyO tyS true lx' γ t () #(Z_of_rwlock_st st2))%I.
    set R : time → () → lit → vProp Σ := (λ _ _ _, True)%I.
    iDestruct "R" as (tR vR) "R".
    iApply (GPS_aSP_SharedWriter_CAS_int_strong rwlockN
              (rwlock_interp γs β tyO tyS) rwlock_noCAS _ β qβ q true _ _ _ _
              #(Z_of_rwlock_st st2 - 1) tR () _ _ P Q Q1 Q2 R _ ∅ with
              "[$LFT $Hβ $R H◯ Hν]");
      [solve_ndisj|done|done|done|set_solver|by left|by left|by right|..].
    { iDestruct "R'" as "[R' AT]". iSplitL""; first by iFrame "AT". iSplitL"".
      { iIntros "!> !>" (????). by iApply rwlock_interp_comparable. }
      { simpl. iSplitR "H◯ Hν"; last by iFrame "∗".
        iIntros (t2 [] [_ Ext2]). iSplit; last first.
        { iIntros "!>" (v ?) "!>". by iSplit; iIntros "$". } iSplitL "".
        { iIntros "!>". iDestruct 1 as (st') "(>#Eq & #Share & F)".
          iIntros "!>". iSplitL ""; iNext; iExists st'; by iFrame "Eq Share". }
        iIntros "(H◯ & Hν)". iDestruct 1 as (st2') "(>Eqst & #Share & >H● & INT)".
        iDestruct "Eqst" as %Eqst.
        iDestruct (rwown_global_reading_st with "H● H◯") as %[q' [n [Eqst2' CASE]]].
        subst st2'. simpl in Eqst. rewrite Eqst /=. iDestruct "INT" as "[$ INT]".
        iExists (). iSplitL ""; [done|].
        iDestruct "INT" as (qr) "(_ & Hh & #Hshr & >Hqrq' & >Hqr & Rm)".
        iDestruct "Hqrq'" as %Hqrq'. iCombine "Hν" "Hqr" as "Hν".
        iIntros "!>" (t3 Ext3) "W R !>". iSplitL "".
        { rewrite /Q1 Eqst. by iIntros "!> $". }
        destruct CASE as [[??]|[[q0 ?] Eqn]]; [subst q' n|subst q'].
        - rewrite Hqrq'. iMod (rwown_update_read_dealloc_1 with "H● H◯") as "H●".
          iApply (step_fupd_mask_mono (↑lftN ∪ ↑histN) _ _ (↑histN));
            [solve_ndisj|apply union_least; solve_ndisj|..].
          iMod ("H†" with "Hν") as "Hν". iIntros "!> !>". iMod "Hν".
          iMod ("Hh" with "Hν") as "Hb". iModIntro.
          inversion Eqst as [Eqst1]. rewrite Eqst1 /=. iSplitL""; [done|].
          iExists None. iSplitL ""; [done|]. iFrame "Share H● Hb W".
          iDestruct (GPS_SWSharedReaders_join with "R Rm") as "R".
          by rewrite Hqrq'.
        - inversion Eqst as [Eqst1]. rewrite Eqn Qp_plus_comm Eqst1.
          iMod (rwown_update_read_dealloc with "H● H◯") as "H●".
          iApply step_fupd_intro; [solve_ndisj|]. iIntros "!>". iSplitL ""; [done|].
          have Eqn': (Z.pos n - 1) = Z.pos (Pos.pred n).
          { rewrite Pos2Z.inj_pred => // ?. by subst n. } rewrite Eqn'.
          iExists (Some (inr (ν, q0, Pos.pred n))). iSplitL ""; [done|].
          iFrame "Share H● W". iExists (q + qr)%Qp. iFrame "H† Hh Hshr Hν".
          iSplitR "R Rm"; [by rewrite Qp_plus_assoc (Qp_plus_comm q0)|].
          by iApply (GPS_SWSharedReaders_join with "R Rm"). } }

    iIntros "!>" (b t [] v') "(Hβ & _ & CASE)".
    iDestruct "CASE" as "[[[% [% %]] _]|[[% _] [R'' [_ [H◯ Hν]]]]]".
    - subst b v'. iMod ("Hcloseα" with "Hβ") as "Hα".
      iMod ("Hclose" with "Hα HL") as "HL". wp_if.
      iApply (type_type _ _ _ [ x ◁ box (uninit 1)]
              with "[] LFT HE Hna HL Hk"); first last.
      { rewrite tctx_interp_singleton tctx_hasty_val //. }
      iApply type_delete; [solve_typing..|].
      iApply type_new; [solve_typing..|]. iIntros (r). simpl_subst.
      iApply type_jump; solve_typing.
    - subst b. iMod ("Hcloseα" with "Hβ") as "Hα".
      iMod ("Hclose" with "Hα HL") as "HL". wp_if.
      iApply (type_type _ _ _ [ x ◁ box (uninit 1); #lx' ◁ rwlockreadguard α ty]
              with "[] LFT HE Hna HL Hk"); first last.
      { rewrite tctx_interp_cons tctx_interp_singleton tctx_hasty_val
                tctx_hasty_val' //=; simpl.
        iFrame "Hx". iExists _,_,_,_. iFrame "Hx' Hαβ Hν H◯ H†".
        iExists _,_,_. iSplitL "R''".
        - iExists _,_; by iDestruct "R''" as "[$ ?]".
        - iFrame "EqO EqS". iExists _,_. by iFrame "R'". }
      iApply type_jump; solve_typing.
  Qed.
End rwlockreadguard_functions.
