From Coq.QArith Require Import Qcanon.
From iris.bi Require Import fractional.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import auth csum frac agree.
From gpfsl.gps Require Import middleware.
From lrust.logic Require Import gps.
From lrust.lifetime Require Import at_borrow.
From lrust.typing Require Import typing.
From lrust.typing.lib.rwlock Require Import rwlock rwlockwriteguard.
Set Default Proof Using "Type".

Section rwlockwriteguard_functions.
  Context `{!typeG Σ, !rwlockG Σ}.

  (* Turning a rwlockwriteguard into a shared borrow. *)
  Definition rwlockwriteguard_deref : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
      let: "r'" := !"x'" +ₗ #1 in
      letalloc: "r" <- "r'" in
      delete [ #1; "x"];; return: ["r"].

  Lemma rwlockwriteguard_deref_type ty `{!TyWf ty} :
    typed_val rwlockwriteguard_deref
      (fn(∀ '(α, β), ∅; &shr{α}(rwlockwriteguard β ty)) → &shr{α} ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros ([α β] ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x').
    iIntros (tid) "#LFT #HE Hna HL Hk HT". simpl_subst.
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']". destruct x' as [[|lx'|]|]; try done.
    iDestruct "Hx'" as (l') "#[Hfrac Hshr]".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "([Hα1 Hα2] & HL & Hclose)";
      [solve_typing..|].
    iMod (frac_bor_acc with "LFT Hfrac Hα1") as (qlx') "[H↦ Hcloseα1]". done.
    rewrite heap_mapsto_vec_singleton.
    iMod (lctx_lft_alive_tok β with "HE HL") as (qβ) "(Hβ & HL & Hclose')";
      [solve_typing..|].
    iDestruct (lft_intersect_acc with "Hβ Hα2") as (qβα) "[Hα2β Hcloseβα2]".
    wp_bind (!(LitV lx'))%E. iApply (wp_step_fupd with "[Hα2β]");
         [done| |by iApply ("Hshr" with "[] Hα2β")|]; first done.
    iMod "H↦" as "[H↦1 H↦2]". wp_read. iIntros "[#Hshr' Hα2β]!>". wp_op. wp_let.
    iDestruct ("Hcloseβα2" with "Hα2β") as "[Hβ Hα2]".
    iMod ("Hcloseα1" with "[$H↦1 $H↦2]") as "Hα1". iMod ("Hclose'" with "Hβ HL") as "HL".
    iMod ("Hclose" with "[$] HL") as "HL".
    iDestruct (lctx_lft_incl_incl α β with "HL HE") as "#Hαβ"; [solve_typing..|].
    iApply (type_type _ _ _ [ x ◁ box (&shr{α}(rwlockwriteguard β ty));
                              #(l' +ₗ 1) ◁ &shr{α}ty]
            with "[] LFT HE Hna HL Hk"); last first.
    { rewrite tctx_interp_cons tctx_interp_singleton tctx_hasty_val tctx_hasty_val' //.
      iFrame. iApply (ty_shr_mono with "[] Hshr'"). iApply lft_incl_glb; first done.
        by iApply lft_incl_refl. }
    iApply (type_letalloc_1 (&shr{α}ty)); [solve_typing..|].
    iIntros (r). simpl_subst. iApply type_delete; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  (* Turning a rwlockwriteguard into a unique borrow. *)
  Definition rwlockwriteguard_derefmut : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
      let: "r'" := !"x'" +ₗ #1 in
      letalloc: "r" <- "r'" in
      delete [ #1; "x"];; return: ["r"].

  Lemma rwlockwriteguard_derefmut_type ty `{!TyWf ty} :
    typed_val rwlockwriteguard_derefmut
      (fn(∀ '(α, β), ∅; &uniq{α}(rwlockwriteguard β ty)) → &uniq{α}ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros ([α β] ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x').
    iIntros (tid) "#LFT #HE Hna HL Hk HT". simpl_subst.
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']". destruct x' as [[|lx'|]|]; try done.
    iMod (bor_exists with "LFT Hx'") as (vl) "H". done.
    iMod (bor_sep with "LFT H") as "[H↦ H]". done.
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)"; [solve_typing..|].
    destruct vl as [|[[|l|]|][]];
      try by iMod (bor_persistent with "LFT H Hα") as "[>[] _]".
    rewrite heap_mapsto_vec_singleton.
    iMod (bor_exists with "LFT H") as (γ) "H". done.
    iMod (bor_exists with "LFT H") as (δ) "H". done.
    iMod (bor_sep with "LFT H") as "[Hb H]". done.
    iMod (bor_sep with "LFT H") as "[Hβδ _]". done.
    iMod (bor_persistent with "LFT Hβδ Hα") as "[#Hβδ Hα]". done.
    iMod (bor_acc with "LFT H↦ Hα") as "[H↦ Hcloseα]". done.
    wp_bind (!(LitV lx'))%E. iMod (bor_unnest with "LFT Hb") as "Hb"; first done.
    wp_read. wp_op. wp_let. iMod "Hb".
    iMod ("Hcloseα" with "[$H↦]") as "[_ Hα]". iMod ("Hclose" with "Hα HL") as "HL".
    iDestruct (lctx_lft_incl_incl α β with "HL HE") as "#Hαβ"; [solve_typing..|].
    iApply (type_type _ _ _ [ x ◁ box (uninit 1); #(l +ₗ 1) ◁ &uniq{α}ty]
            with "[] LFT HE Hna HL Hk"); last first.
    { rewrite tctx_interp_cons tctx_interp_singleton tctx_hasty_val tctx_hasty_val' //.
      iFrame. iApply (bor_shorten with "[] Hb"). iApply lft_incl_glb.
      by iApply lft_incl_trans. by iApply lft_incl_refl. }
    iApply (type_letalloc_1 (&uniq{α}ty)); [solve_typing..|].
    iIntros (r). simpl_subst. iApply type_delete; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  (* Drop a rwlockwriteguard and release the lock. *)
  Definition rwlockwriteguard_drop : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
      "x'" <-ʳᵉˡ #0;;
      delete [ #1; "x"];;
      let: "r" := new [ #0] in return: ["r"].

  Lemma rwlockwriteguard_drop_type ty `{!TyWf ty} :
    typed_val rwlockwriteguard_drop
              (fn(∀ α, ∅; rwlockwriteguard α ty) → unit).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (α ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
    iIntros (tid) "#LFT #HE Hna HL Hk HT".
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']".
    destruct x' as [[|lx'|]|]; try done. simpl.
    iDestruct "Hx'" as (γs β) "(Hx' & #Hαβ & wst & Hinv)".
    iDestruct "Hinv" as (γ tyO tyS) "(W & #EqO & #EqS & #R)".
    iDestruct "W" as (t) "W".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)";
      [solve_typing..|].
    iMod (lft_incl_acc with "Hαβ Hα") as (qβ) "[Hβ Hcloseα]". done.
    wp_bind (#lx' <-ʳᵉˡ #0)%E.
    iAssert (▷ □ (∀ κ q' E', ⌜lftE ⊆ E'⌝ -∗ &{κ} tyO -∗
                      (q').[κ] ={E'}=∗ (□ tyS κ) ∗ (q').[κ]))%I as "#Share".
    { iIntros "!> !#" (κ???) "tyO tok".
      iMod (ty_share with "LFT [tyO] tok") as "[#tyS $]" ;[done|..].
      - iApply (bor_iff with "EqO tyO").
      - iIntros "!> !#". by iApply "EqS". }
    iApply (GPS_aSP_SWWrite_rel rwlockN (rwlock_interp γs β tyO tyS)
              rwlock_noCAS _ β qβ (λ _, True)%I True%I
              (rwlock_interp γs β tyO tyS true lx' γ t () #(-1))
              t () () #(-1) #0 with "[$LFT $Hβ $W wst Hx']");
      [solve_ndisj|done..|iSplitL ""; last iSplitR ""|].
    { iDestruct "R" as (??) "[R $]". }
    { iIntros "!>" (t' Ext') "W". iDestruct 1 as (st' Eqv) "[_ [g INT]]".
      iDestruct (rwown_global_writing_st with "g wst") as %?. subst st'.
      iMod (rwown_update_write_dealloc with "g wst") as "g".
      iModIntro. iSplitL ""; [by iIntros "!> _"|].
      iDestruct (GPS_SWWriter_shared with "W") as "[W Rs]".
      iSplitR ""; [|done]. iExists None. iSplitL ""; [done|].
      iFrame "Share g W Rs". iApply (bor_iff with "[] Hx'").
      iIntros "!> !#". iApply bi_iff_sym. by iApply "EqO". }
    { by iIntros "!> !> $". }
    iIntros "!>" (?) "(_ & R' & Hβ & _)".
    wp_seq. iMod ("Hcloseα" with "Hβ") as "Hα".
    iMod ("Hclose" with "Hα HL") as "HL".
    iApply (type_type _ _ _ [ x ◁ box (uninit 1)]
            with "[] LFT HE Hna HL Hk"); first last.
    { rewrite tctx_interp_singleton tctx_hasty_val //. }
    iApply type_delete; [solve_typing..|].
    iApply type_new; [solve_typing..|]. iIntros (r). simpl_subst.
    iApply type_jump; solve_typing.
  Qed.
End rwlockwriteguard_functions.
