From iris.algebra Require Import frac.
From iris.program_logic Require Import adequacy.
From gpfsl.base_logic Require Import adequacy.
From gpfsl.logic Require Import na_invariants.
From lrust.lang Require Import notation.
From lrust.lifetime Require Import lifetime frac_borrow.
From lrust.typing Require Import typing.

Set Default Proof Using "Type".

Class typePreG Σ := PreTypeG {
  type_heapG :> noprolPreG Σ;
  type_lftG :> lftPreG view_Lat Σ;
  type_preG_na_invG :> na_invG view_Lat Σ;
  type_frac_borrowG :> frac_borG Σ
}.

Definition typeΣ : gFunctors :=
  #[noprolΣ; lftΣ view_Lat; na_invΣ view_Lat; GFunctor (constRF fracR)].
Instance subG_typePreG {Σ} : subG typeΣ Σ → typePreG Σ.
Proof. solve_inG. Qed.

Section type_soundness.
  Definition exit_cont : val := λ: [<>], #☠.

  Definition main_type `{!typeG Σ} := (fn(∅) → unit)%T.

  Theorem type_soundness `{!typePreG Σ} (main : val) σ t :
    (∀ `{!typeG Σ}, typed_val main main_type) →
    rtc erased_step ([(main [exit_cont] at init_tview)%E], mkGB ∅ ∅ ∅) (t, σ) →
    (* nonracing_threadpool t σ ∧ *)   (* TODO : prove DRF theorem *)
    (∀ e 𝓥, (e at 𝓥)%E ∈ t → is_Some (to_val e) ∨ reducible (e at 𝓥) σ).
  Proof.
    intros Hmain Hrtc.
    cut (adequate NotStuck (main [exit_cont]%E at init_tview) (mkGB ∅ ∅ ∅) (λ _ _, True)).
    { (* split. by eapply adequate_nonracing. *)
      intros.
      edestruct (adequate_not_stuck NotStuck (main [exit_cont]%E at init_tview))
         as [toval|red]=>//.
      - destruct toval as [v ?%nopro_lang.to_base_val]. eauto.
      - eauto. }
    apply: (noprol_adequacy' _ _ (λ x, True))=>? tid1.
    iMod (lft_init _ (↑histN)) as (?) "#LFT"; [done|solve_ndisj|].
    iMod na_alloc as (tid2) "Htl". set (Htype := TypeG _ _ _ _ _).
    set (tid := {| tid_na_inv_pool := tid2; tid_tid := tid1 |}).
    wp_bind (of_val main). iApply (wp_wand with "[Htl]").
    { iApply (Hmain Htype [] [] $! tid with "LFT [] Htl [] []");
        by rewrite /elctx_interp /llctx_interp ?tctx_interp_nil. }
    clear Hrtc Hmain main. iIntros (main) "(Htl & _ & Hmain & _)".
    iDestruct "Hmain" as (?) "[EQ Hmain]".
    rewrite eval_path_of_val. iDestruct "EQ" as %[= <-].
    iDestruct "Hmain" as (f k x e ?) "(EQ & % & Hmain)". iDestruct "EQ" as %[= ->].
    destruct x; try done. wp_rec.
    iMod (lft_create with "LFT") as (ϝ) "Hϝ"; first done.
    iApply ("Hmain" $! () ϝ exit_cont [#] tid with "LFT [] Htl [Hϝ] []");
      last by rewrite tctx_interp_nil.
    { by rewrite /elctx_interp /=. }
    { rewrite /llctx_interp /=. iSplit; last done. iExists ϝ. iFrame. by rewrite /= left_id. }
    rewrite cctx_interp_singleton. simpl. iIntros (args) "_ _".
    inv_vec args. iIntros (x) "_ /=". by wp_lam.
  Qed.
End type_soundness.

(* Soundness theorem when no other CMRA is needed. *)

Theorem type_soundness_closed (main : val) σ t :
  (∀ `{!typeG typeΣ}, typed_val main main_type) →
  rtc erased_step ([(main [exit_cont] at init_tview)%E], mkGB ∅ ∅ ∅) (t, σ) →
  (* nonracing_threadpool t σ ∧ *)   (* TODO : prove DRF theorem *)
  (∀ e 𝓥, (e at 𝓥)%E ∈ t → is_Some (to_val e) ∨ reducible (e at 𝓥) σ).
Proof.
  intros. eapply @type_soundness; try done. apply _.
Qed.
